import functools
import inspect
import warnings
from datetime import datetime, timedelta
from typing import Tuple


def deprecated(reason):
    """This is a decorator which can be used to mark functions and classes as
    deprecated. It will result in a warning being emitted when the function is used.

    From https://stackoverflow.com/a/40301488
    """

    string_types = (type(b""), type(u""))
    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                warnings.simplefilter("default", DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):  # noqa: SIM106

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))


def pretty_seconds(x: float) -> timedelta:
    """Return a timedelta object from seconds."""
    return timedelta(seconds=x)


def pretty_time(x: float) -> datetime:
    """Return a datetime object from unix timestamp."""
    return datetime.fromtimestamp(x)


def int_to_rgb(x: int) -> Tuple[int, int, int]:
    """Return a RGB tuple from integer."""
    red = (x >> 16) & 0xFF
    green = (x >> 8) & 0xFF
    blue = x & 0xFF
    return red, green, blue


def rgb_to_int(x: Tuple[int, int, int]) -> int:
    """Return an integer from RGB tuple."""
    return int(x[0] << 16 | x[1] << 8 | x[2])


def int_to_brightness(x: int) -> int:
    """"Return brightness (0-100) from integer."""
    return x >> 24


def brightness_and_color_to_int(brightness: int, color: Tuple[int, int, int]) -> int:
    return int(brightness << 24 | color[0] << 16 | color[1] << 8 | color[2])



import base64
import json
import os
from unittest import TestCase

import pytest

from miio import ChuangmiIr
from miio.chuangmi_ir import ChuangmiIrException

from .dummies import DummyDevice

with open(os.path.join(os.path.dirname(__file__), "test_chuangmi_ir.json")) as inp:
    test_data = json.load(inp)


class DummyChuangmiIr(DummyDevice, ChuangmiIr):
    def __init__(self, *args, **kwargs):
        self.state = {"last_ir_played": None}
        self.return_values = {
            "miIO.ir_learn": lambda x: True,
            "miIO.ir_read": lambda x: True,
            "miIO.ir_play": self._ir_play_input_validation,
        }
        super().__init__(args, kwargs)

    def _ir_play_input_validation(self, payload):
        try:
            base64.b64decode(payload["code"])
            self._set_state("last_ir_played", [[payload["code"], payload.get("freq")]])
            return True
        except TypeError:
            return False


@pytest.fixture(scope="class")
def chuangmiir(request):
    request.cls.device = DummyChuangmiIr()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("chuangmiir")
class TestChuangmiIr(TestCase):
    def test_learn(self):
        assert self.device.learn() is True
        assert self.device.learn(30) is True

        with pytest.raises(ChuangmiIrException):
            self.device.learn(-1)

        with pytest.raises(ChuangmiIrException):
            self.device.learn(1000001)

    def test_read(self):
        assert self.device.read() is True
        assert self.device.read(30) is True

        with pytest.raises(ChuangmiIrException):
            self.device.read(-1)

        with pytest.raises(ChuangmiIrException):
            self.device.read(1000001)

    def test_play_raw(self):
        for args in test_data["test_raw_ok"]:
            with self.subTest():
                self.device._reset_state()
                self.assertTrue(self.device.play_raw(*args["in"]))
                self.assertSequenceEqual(
                    self.device.state["last_ir_played"], args["out"]
                )

    def test_pronto_to_raw(self):
        for args in test_data["test_pronto_ok"]:
            with self.subTest():
                self.assertSequenceEqual(
                    ChuangmiIr.pronto_to_raw(*args["in"]), args["out"]
                )

        for args in test_data["test_pronto_exception"]:
            with self.subTest(), pytest.raises(ChuangmiIrException):
                ChuangmiIr.pronto_to_raw(*args["in"])

    def test_play_pronto(self):
        for args in test_data["test_pronto_ok"]:
            with self.subTest():
                self.device._reset_state()
                self.assertTrue(self.device.play_pronto(*args["in"]))
                self.assertSequenceEqual(
                    self.device.state["last_ir_played"], args["out"]
                )

        for args in test_data["test_pronto_exception"]:
            with pytest.raises(ChuangmiIrException):
                self.device.play_pronto(*args["in"])

    def test_play_auto(self):
        for args in test_data["test_raw_ok"] + test_data["test_pronto_ok"]:
            if len(args["in"]) > 1:  # autodetect doesn't take any extra args
                continue
            with self.subTest():
                self.device._reset_state()
                self.assertTrue(self.device.play(*args["in"]))
                self.assertSequenceEqual(
                    self.device.state["last_ir_played"], args["out"]
                )

    def test_play_with_type(self):
        for type_, tests in [
            ("raw", test_data["test_raw_ok"]),
            ("pronto", test_data["test_pronto_ok"]),
        ]:
            for args in tests:
                with self.subTest():
                    command = "{}:{}".format(type_, ":".join(map(str, args["in"])))
                    self.assertTrue(self.device.play(command))
                    self.assertSequenceEqual(
                        self.device.state["last_ir_played"], args["out"]
                    )
        with pytest.raises(ChuangmiIrException):
            self.device.play("invalid:command")

        with pytest.raises(ChuangmiIrException):
            self.device.play("pronto:command:invalid:argument:count")

        with pytest.raises(ChuangmiIrException):
            self.device.play("pronto:command:invalidargument")



from unittest import TestCase

import pytest

from miio.airfilter_util import FilterType, FilterTypeUtil


@pytest.fixture(scope="class")
def airfilter_util(request):
    request.cls.filter_type_util = FilterTypeUtil()


@pytest.mark.usefixtures("airfilter_util")
class TestAirFilterUtil(TestCase):
    def test_determine_filter_type__recognises_unknown_filter(self):
        assert (
            self.filter_type_util.determine_filter_type("0:0:0:0:0:0:0", None)
            is FilterType.Unknown
        )

    def test_determine_filter_type__recognises_antibacterial_filter(self):
        assert (
            self.filter_type_util.determine_filter_type(
                "80:64:d1:ba:4f:5f:4", "12:34:41:30"
            )
            is FilterType.AntiBacterial
        )

    def test_determine_filter_type__recognises_antiformaldehyde_filter(self):
        assert (
            self.filter_type_util.determine_filter_type(
                "80:64:d1:ba:4f:5f:4", "12:34:00:31"
            )
            is FilterType.AntiFormaldehyde
        )

    def test_determine_filter_type__falls_back_to_regular_filter(self):
        regular_filters = [
            "12:34:56:78",
            "12:34:56:31",
            "12:34:56:31:11:11",
            "CO:FF:FF:EE",
            None,
        ]
        for product_id in regular_filters:
            assert (
                self.filter_type_util.determine_filter_type(
                    "80:64:d1:ba:4f:5f:4", product_id
                )
                is FilterType.Regular
            )



"""Vacuum 1C STYTJ01ZHM (dreame.vacuum.mc1808)"""

import logging
from enum import Enum

from .click_common import command, format_output
from .miot_device import DeviceStatus as DeviceStatusContainer
from .miot_device import MiotDevice, MiotMapping

_LOGGER = logging.getLogger(__name__)

_MAPPING: MiotMapping = {
    "battery_level": {"siid": 2, "piid": 1},
    "charging_state": {"siid": 2, "piid": 2},
    "device_fault": {"siid": 3, "piid": 1},
    "device_status": {"siid": 3, "piid": 2},
    "brush_left_time": {"siid": 26, "piid": 1},
    "brush_life_level": {"siid": 26, "piid": 2},
    "filter_life_level": {"siid": 27, "piid": 1},
    "filter_left_time": {"siid": 27, "piid": 2},
    "brush_left_time2": {"siid": 28, "piid": 1},
    "brush_life_level2": {"siid": 28, "piid": 2},
    "operating_mode": {"siid": 18, "piid": 1},
    "cleaning_mode": {"siid": 18, "piid": 6},
    "delete_timer": {"siid": 18, "piid": 8},
    "life_sieve": {"siid": 19, "piid": 1},
    "life_brush_side": {"siid": 19, "piid": 2},
    "life_brush_main": {"siid": 19, "piid": 3},
    "timer_enable": {"siid": 20, "piid": 1},
    "start_time": {"siid": 20, "piid": 2},
    "stop_time": {"siid": 20, "piid": 3},
    "deg": {"siid": 21, "piid": 1, "access": ["write"]},
    "speed": {"siid": 21, "piid": 2, "access": ["write"]},
    "map_view": {"siid": 23, "piid": 1},
    "frame_info": {"siid": 23, "piid": 2},
    "volume": {"siid": 24, "piid": 1},
    "voice_package": {"siid": 24, "piid": 3},
}


class ChargingState(Enum):
    Unknown = -1
    Charging = 1
    Discharging = 2
    Charging2 = 4
    GoCharging = 5


class CleaningMode(Enum):
    Unknown = -1
    Quiet = 0
    Default = 1
    Medium = 2
    Strong = 3


class OperatingMode(Enum):
    Unknown = -1
    Paused = 1
    Cleaning = 2
    GoCharging = 3
    Charging = 6
    ManualCleaning = 13
    Sleeping = 14
    ManualPaused = 17
    ZonedCleaning = 19


class FaultStatus(Enum):
    Unknown = -1
    NoFaults = 0


class DeviceStatus(Enum):
    Unknown = -1
    Sweeping = 1
    Idle = 2
    Paused = 3
    Error = 4
    GoCharging = 5
    Charging = 6
    ManualSweeping = 13


class DreameVacuumStatus(DeviceStatusContainer):
    def __init__(self, data):
        self.data = data

    @property
    def battery_level(self) -> str:
        return self.data["battery_level"]

    @property
    def brush_left_time(self) -> str:
        return self.data["brush_left_time"]

    @property
    def brush_left_time2(self) -> str:
        return self.data["brush_left_time2"]

    @property
    def brush_life_level2(self) -> str:
        return self.data["brush_life_level2"]

    @property
    def brush_life_level(self) -> str:
        return self.data["brush_life_level"]

    @property
    def filter_left_time(self) -> str:
        return self.data["filter_left_time"]

    @property
    def filter_life_level(self) -> str:
        return self.data["filter_life_level"]

    @property
    def device_fault(self) -> FaultStatus:
        try:
            return FaultStatus(self.data["device_fault"])
        except ValueError:
            _LOGGER.error("Unknown FaultStatus (%s)", self.data["device_fault"])
            return FaultStatus.Unknown

    @property
    def charging_state(self) -> ChargingState:
        try:
            return ChargingState(self.data["charging_state"])
        except ValueError:
            _LOGGER.error("Unknown ChargingStats (%s)", self.data["charging_state"])
            return ChargingState.Unknown

    @property
    def operating_mode(self) -> OperatingMode:
        try:
            return OperatingMode(self.data["operating_mode"])
        except ValueError:
            _LOGGER.error("Unknown OperatingMode (%s)", self.data["operating_mode"])
            return OperatingMode.Unknown

    @property
    def cleaning_mode(self) -> CleaningMode:
        try:
            return CleaningMode(self.data["cleaning_mode"])
        except ValueError:
            _LOGGER.error("Unknown CleaningMode (%s)", self.data["cleaning_mode"])
            return CleaningMode.Unknown

    @property
    def device_status(self) -> DeviceStatus:
        try:
            return DeviceStatus(self.data["device_status"])
        except TypeError:
            _LOGGER.error("Unknown DeviceStatus (%s)", self.data["device_status"])
            return DeviceStatus.Unknown

    @property
    def life_sieve(self) -> str:
        return self.data["life_sieve"]

    @property
    def life_brush_side(self) -> str:
        return self.data["life_brush_side"]

    @property
    def life_brush_main(self) -> str:
        return self.data["life_brush_main"]

    @property
    def timer_enable(self) -> str:
        return self.data["timer_enable"]

    @property
    def start_time(self) -> str:
        return self.data["start_time"]

    @property
    def stop_time(self) -> str:
        return self.data["stop_time"]

    @property
    def map_view(self) -> str:
        return self.data["map_view"]

    @property
    def volume(self) -> str:
        return self.data["volume"]

    @property
    def voice_package(self) -> str:
        return self.data["voice_package"]


class DreameVacuumMiot(MiotDevice):
    """Interface for Vacuum 1C STYTJ01ZHM (dreame.vacuum.mc1808)"""

    mapping = _MAPPING

    @command(
        default_output=format_output(
            "\n",
            "Battery level: {result.battery_level}\n"
            "Brush life level: {result.brush_life_level}\n"
            "Brush left time: {result.brush_left_time}\n"
            "Charging state: {result.charging_state.name}\n"
            "Cleaning mode: {result.cleaning_mode.name}\n"
            "Device fault: {result.device_fault.name}\n"
            "Device status: {result.device_status.name}\n"
            "Filter left level: {result.filter_left_time}\n"
            "Filter life level: {result.filter_life_level}\n"
            "Life brush main: {result.life_brush_main}\n"
            "Life brush side: {result.life_brush_side}\n"
            "Life sieve: {result.life_sieve}\n"
            "Map view: {result.map_view}\n"
            "Operating mode: {result.operating_mode.name}\n"
            "Side cleaning brush left time: {result.brush_left_time2}\n"
            "Side cleaning brush life level: {result.brush_life_level2}\n"
            "Timer enabled: {result.timer_enable}\n"
            "Timer start time: {result.start_time}\n"
            "Timer stop time: {result.stop_time}\n"
            "Voice package: {result.voice_package}\n"
            "Volume: {result.volume}\n",
        )
    )
    def status(self) -> DreameVacuumStatus:
        """State of the vacuum."""

        return DreameVacuumStatus(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping(max_properties=10)
            }
        )

    def send_action(self, siid, aiid, params=None):
        """Send action to device."""

        # {"did":"<mydeviceID>","siid":18,"aiid":1,"in":[{"piid":1,"value":2}]
        if params is None:
            params = []
        payload = {
            "did": f"call-{siid}-{aiid}",
            "siid": siid,
            "aiid": aiid,
            "in": params,
        }
        return self.send("action", payload)

    @command()
    def start(self) -> None:
        """Start cleaning."""
        return self.send_action(3, 1)

    @command()
    def stop(self) -> None:
        """Stop cleaning."""
        return self.send_action(3, 2)

    @command()
    def home(self) -> None:
        """Return to home."""
        return self.send_action(2, 1)

    @command()
    def identify(self) -> None:
        """Locate the device (i am here)."""
        return self.send_action(17, 1)

    @command()
    def reset_mainbrush_life(self) -> None:
        """Reset main brush life."""
        return self.send_action(26, 1)

    @command()
    def reset_filter_life(self) -> None:
        """Reset filter life."""
        return self.send_action(27, 1)

    @command()
    def reset_sidebrush_life(self) -> None:
        """Reset side brush life."""
        return self.send_action(28, 1)



from miio.click_common import validate_ip, validate_token


def test_validate_token_empty():
    assert not validate_token(None, None, None)


def test_validate_ip_empty():
    assert validate_ip(None, None, None) is None



from unittest import TestCase

import pytest

from miio import AirQualityMonitorCGDN1
from miio.airqualitymonitor_miot import (
    AirQualityMonitorMiotException,
    ChargingState,
    DisplayTemperatureUnitCGDN1,
)

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "humidity": 34,
    "pm25": 10,
    "pm10": 15,
    "temperature": 18.599999,
    "co2": 620,
    "battery": 20,
    "charging_state": 2,
    "voltage": 26,
    "start_time": 0,
    "end_time": 0,
    "monitoring_frequency": 1,
    "screen_off": 15,
    "device_off": 30,
    "temperature_unit": "c",
}


class DummyAirQualityMonitorCGDN1(DummyMiotDevice, AirQualityMonitorCGDN1):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        self.return_values = {
            "get_prop": self._get_state,
            "set_monitoring_frequency": lambda x: self._set_state(
                "monitoring_frequency", x
            ),
            "set_device_off_duration": lambda x: self._set_state("device_off", x),
            "set_screen_off_duration": lambda x: self._set_state("screen_off", x),
            "set_display_temperature_unit": lambda x: self._set_state(
                "temperature_unit", x
            ),
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def airqualitymonitorcgdn1(request):
    request.cls.device = DummyAirQualityMonitorCGDN1()


@pytest.mark.usefixtures("airqualitymonitorcgdn1")
class TestAirQualityMonitor(TestCase):
    def test_status(self):
        status = self.device.status()
        assert status.humidity is _INITIAL_STATE["humidity"]
        assert status.pm25 is _INITIAL_STATE["pm25"]
        assert status.pm10 is _INITIAL_STATE["pm10"]
        assert status.temperature is _INITIAL_STATE["temperature"]
        assert status.co2 is _INITIAL_STATE["co2"]
        assert status.battery is _INITIAL_STATE["battery"]
        assert status.charging_state is ChargingState(_INITIAL_STATE["charging_state"])
        assert status.monitoring_frequency is _INITIAL_STATE["monitoring_frequency"]
        assert status.screen_off is _INITIAL_STATE["screen_off"]
        assert status.device_off is _INITIAL_STATE["device_off"]
        assert status.display_temperature_unit is DisplayTemperatureUnitCGDN1(
            _INITIAL_STATE["temperature_unit"]
        )

    def test_set_monitoring_frequency_duration(self):
        def monitoring_frequency():
            return self.device.status().monitoring_frequency

        self.device.set_monitoring_frequency_duration(0)
        assert monitoring_frequency() == 0

        self.device.set_monitoring_frequency_duration(290)
        assert monitoring_frequency() == 290

        self.device.set_monitoring_frequency_duration(600)
        assert monitoring_frequency() == 600

        with pytest.raises(AirQualityMonitorMiotException):
            self.device.set_monitoring_frequency_duration(-1)

        with pytest.raises(AirQualityMonitorMiotException):
            self.device.set_monitoring_frequency_duration(601)

    def test_set_device_off_duration(self):
        def device_off_duration():
            return self.device.status().device_off

        self.device.set_device_off_duration(0)
        assert device_off_duration() == 0

        self.device.set_device_off_duration(29)
        assert device_off_duration() == 29

        self.device.set_device_off_duration(60)
        assert device_off_duration() == 60

        with pytest.raises(AirQualityMonitorMiotException):
            self.device.set_device_off_duration(-1)

        with pytest.raises(AirQualityMonitorMiotException):
            self.device.set_device_off_duration(61)

    def test_set_screen_off_duration(self):
        def screen_off_duration():
            return self.device.status().screen_off

        self.device.set_screen_off_duration(0)
        assert screen_off_duration() == 0

        self.device.set_screen_off_duration(140)
        assert screen_off_duration() == 140

        self.device.set_screen_off_duration(300)
        assert screen_off_duration() == 300

        with pytest.raises(AirQualityMonitorMiotException):
            self.device.set_screen_off_duration(-1)

        with pytest.raises(AirQualityMonitorMiotException):
            self.device.set_screen_off_duration(301)

    def test_set_display_temperature_unit(self):
        def display_temperature_unit():
            return self.device.status().display_temperature_unit

        self.device.set_display_temperature_unit(DisplayTemperatureUnitCGDN1.Celcius)
        assert display_temperature_unit() == DisplayTemperatureUnitCGDN1.Celcius

        self.device.set_display_temperature_unit(DisplayTemperatureUnitCGDN1.Fahrenheit)
        assert display_temperature_unit() == DisplayTemperatureUnitCGDN1.Fahrenheit



"""Xiaomi Zigbee lights."""

import click

from ...click_common import command
from .subdevice import SubDevice


class LightBulb(SubDevice):
    """Base class for subdevice light bulbs."""

    @command()
    def on(self):
        """Turn bulb on."""
        return self.send_arg("set_power", ["on"]).pop()

    @command()
    def off(self):
        """Turn bulb off."""
        return self.send_arg("set_power", ["off"]).pop()

    @command(click.argument("ctt", type=int))
    def set_color_temp(self, ctt):
        """Set the color temperature of the bulb ctt_min-ctt_max."""
        return self.send_arg("set_ct", [ctt]).pop()

    @command(click.argument("brightness", type=int))
    def set_brightness(self, brightness):
        """Set the brightness of the bulb 1-100."""
        return self.send_arg("set_bright", [brightness]).pop()



import datetime
from unittest import TestCase
from unittest.mock import patch

import pytest

from miio import Vacuum, VacuumStatus
from miio.vacuum import CarpetCleaningMode, MopMode

from .dummies import DummyDevice


class DummyVacuum(DummyDevice, Vacuum):
    STATE_CHARGING = 8
    STATE_CLEANING = 5
    STATE_ZONED_CLEAN = 9
    STATE_IDLE = 3
    STATE_HOME = 6
    STATE_SPOT = 11
    STATE_GOTO = 4
    STATE_ERROR = 12
    STATE_PAUSED = 10
    STATE_MANUAL = 7

    def __init__(self, *args, **kwargs):
        self.state = {
            "state": 8,
            "dnd_enabled": 1,
            "clean_time": 0,
            "msg_ver": 4,
            "map_present": 1,
            "error_code": 0,
            "in_cleaning": 0,
            "clean_area": 0,
            "battery": 100,
            "fan_power": 20,
            "msg_seq": 320,
            "water_box_status": 1,
        }

        self.return_values = {
            "get_status": self.vacuum_state,
            "app_start": lambda x: self.change_mode("start"),
            "app_stop": lambda x: self.change_mode("stop"),
            "app_pause": lambda x: self.change_mode("pause"),
            "app_spot": lambda x: self.change_mode("spot"),
            "app_goto_target": lambda x: self.change_mode("goto"),
            "app_zoned_clean": lambda x: self.change_mode("zoned clean"),
            "app_charge": lambda x: self.change_mode("charge"),
            "miIO.info": "dummy info",
        }

        super().__init__(args, kwargs)
        self.model = None

    def change_mode(self, new_mode):
        if new_mode == "spot":
            self.state["state"] = DummyVacuum.STATE_SPOT
        elif new_mode == "home":
            self.state["state"] = DummyVacuum.STATE_HOME
        elif new_mode == "pause":
            self.state["state"] = DummyVacuum.STATE_PAUSED
        elif new_mode == "start":
            self.state["state"] = DummyVacuum.STATE_CLEANING
        elif new_mode == "stop":
            self.state["state"] = DummyVacuum.STATE_IDLE
        elif new_mode == "goto":
            self.state["state"] = DummyVacuum.STATE_GOTO
        elif new_mode == "zoned clean":
            self.state["state"] = DummyVacuum.STATE_ZONED_CLEAN
        elif new_mode == "charge":
            self.state["state"] = DummyVacuum.STATE_CHARGING

    def vacuum_state(self, _):
        return [self.state]


@pytest.fixture(scope="class")
def dummyvacuum(request):
    request.cls.device = DummyVacuum()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("dummyvacuum")
class TestVacuum(TestCase):
    def status(self):
        return self.device.status()

    def test_status(self):
        self.device._reset_state()

        assert repr(self.status()) == repr(VacuumStatus(self.device.start_state))

        status = self.status()
        assert status.is_on is False
        assert status.clean_time == datetime.timedelta()
        assert status.error_code == 0
        assert status.error == "No error"
        assert status.fanspeed == self.device.start_state["fan_power"]
        assert status.battery == self.device.start_state["battery"]
        assert status.is_water_box_attached is True

    def test_status_with_errors(self):
        errors = {5: "Clean main brush", 19: "Unpowered charging station"}

        for errcode, error in errors.items():
            self.device.state["state"] = self.device.STATE_ERROR
            self.device.state["error_code"] = errcode
            assert self.status().is_on is False
            assert self.status().got_error is True
            assert self.status().error_code == errcode
            assert self.status().error == error

    def test_start_and_stop(self):
        assert self.status().is_on is False
        self.device.start()
        assert self.status().is_on is True
        assert self.status().state_code == self.device.STATE_CLEANING
        self.device.stop()
        assert self.status().is_on is False

    def test_spot(self):
        assert self.status().is_on is False
        self.device.spot()
        assert self.status().is_on is True
        assert self.status().state_code == self.device.STATE_SPOT
        self.device.stop()
        assert self.status().is_on is False

    def test_pause(self):
        self.device.start()
        assert self.status().is_on is True
        self.device.pause()
        assert self.status().state_code == self.device.STATE_PAUSED

    def test_home(self):
        self.device.start()
        assert self.status().is_on is True
        self.device.home()
        assert self.status().state_code == self.device.STATE_CHARGING
        # TODO pause here and update to idle/charging and assert for that?
        # Another option is to mock that app_stop mode is entered before
        # the charging is activated.

    def test_goto(self):
        self.device.start()
        assert self.status().is_on is True
        self.device.goto(24000, 24000)
        assert self.status().state_code == self.device.STATE_GOTO

    def test_zoned_clean(self):
        self.device.start()
        assert self.status().is_on is True
        self.device.zoned_clean(
            [[25000, 25000, 25500, 25500, 3], [23000, 23000, 22500, 22500, 1]]
        )
        assert self.status().state_code == self.device.STATE_ZONED_CLEAN

    def test_timezone(self):
        with patch.object(
            self.device,
            "send",
            return_value=[
                {"olson": "Europe/Berlin", "posix": "CET-1CEST,M3.5.0,M10.5.0/3"}
            ],
        ):
            assert self.device.timezone() == "Europe/Berlin"

        with patch.object(self.device, "send", return_value=["Europe/Berlin"]):
            assert self.device.timezone() == "Europe/Berlin"

        with patch.object(self.device, "send", return_value=0):
            assert self.device.timezone() == "UTC"

    def test_history(self):
        with patch.object(
            self.device,
            "send",
            return_value=[
                174145,
                2410150000,
                82,
                [
                    1488240000,
                    1488153600,
                    1488067200,
                    1487980800,
                    1487894400,
                    1487808000,
                    1487548800,
                ],
            ],
        ):
            assert self.device.clean_history().total_duration == datetime.timedelta(
                days=2, seconds=1345
            )

            assert self.device.clean_history().dust_collection_count is None

            assert self.device.clean_history().ids[0] == 1488240000

    def test_history_dict(self):
        with patch.object(
            self.device,
            "send",
            return_value={
                "clean_time": 174145,
                "clean_area": 2410150000,
                "clean_count": 82,
                "dust_collection_count": 5,
                "records": [
                    1488240000,
                    1488153600,
                    1488067200,
                    1487980800,
                    1487894400,
                    1487808000,
                    1487548800,
                ],
            },
        ):
            assert self.device.clean_history().total_duration == datetime.timedelta(
                days=2, seconds=1345
            )

            assert self.device.clean_history().dust_collection_count == 5

            assert self.device.clean_history().ids[0] == 1488240000

    def test_history_details(self):
        with patch.object(
            self.device,
            "send",
            return_value=[[1488347071, 1488347123, 16, 0, 0, 0]],
        ):
            assert self.device.clean_details(
                123123, return_list=False
            ).duration == datetime.timedelta(seconds=16)

    def test_history_details_dict(self):
        with patch.object(
            self.device,
            "send",
            return_value=[
                {
                    "begin": 1616757243,
                    "end": 1616758193,
                    "duration": 950,
                    "area": 10852500,
                    "error": 0,
                    "complete": 1,
                    "start_type": 2,
                    "clean_type": 1,
                    "finish_reason": 52,
                    "dust_collection_status": 0,
                }
            ],
        ):
            assert self.device.clean_details(
                123123, return_list=False
            ).duration == datetime.timedelta(seconds=950)

    def test_history_empty(self):
        with patch.object(
            self.device,
            "send",
            return_value={
                "clean_time": 174145,
                "clean_area": 2410150000,
                "clean_count": 82,
                "dust_collection_count": 5,
            },
        ):
            assert self.device.clean_history().total_duration == datetime.timedelta(
                days=2, seconds=1345
            )

            assert len(self.device.clean_history().ids) == 0

    def test_carpet_cleaning_mode(self):
        with patch.object(self.device, "send", return_value=[{"carpet_clean_mode": 0}]):
            assert self.device.carpet_cleaning_mode() == CarpetCleaningMode.Avoid

        with patch.object(self.device, "send", return_value="unknown_method"):
            assert self.device.carpet_cleaning_mode() is None

        with patch.object(self.device, "send", return_value=["ok"]) as mock_method:
            assert self.device.set_carpet_cleaning_mode(CarpetCleaningMode.Rise) is True
            mock_method.assert_called_once_with(
                "set_carpet_clean_mode", {"carpet_clean_mode": 1}
            )

    def test_mop_mode(self):
        with patch.object(self.device, "send", return_value=["ok"]) as mock_method:
            assert self.device.set_mop_mode(MopMode.Deep) is True
            mock_method.assert_called_once_with("set_mop_mode", [301])

        with patch.object(self.device, "send", return_value=[300]):
            assert self.device.mop_mode() == MopMode.Standard

        with patch.object(self.device, "send", return_value=[32453]):
            assert self.device.mop_mode() is None



import logging

import click

from .click_common import command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)


class WifiRepeaterException(DeviceException):
    pass


class WifiRepeaterStatus(DeviceStatus):
    def __init__(self, data):
        """
        Response of a xiaomi.repeater.v2:

        {
          'sta': {'count': 2, 'access_policy': 0},
          'mat': [
            {'mac': 'aa:aa:aa:aa:aa:aa', 'ip': '192.168.1.133', 'last_time': 54371873},
            {'mac': 'bb:bb:bb:bb:bb:bb', 'ip': '192.168.1.156', 'last_time': 54371496}
          ],
          'access_list': {'mac': ''}
        }
        """
        self.data = data

    @property
    def access_policy(self) -> int:
        """Access policy of the associated stations."""
        return self.data["sta"]["access_policy"]

    @property
    def associated_stations(self) -> dict:
        """List of associated stations."""
        return self.data["mat"]

    def __repr__(self) -> str:
        s = "<WifiRepeaterStatus access_policy=%s, " "associated_stations=%s>" % (
            self.access_policy,
            len(self.associated_stations),
        )
        return s


class WifiRepeaterConfiguration(DeviceStatus):
    def __init__(self, data):
        """Response of a xiaomi.repeater.v2:

        {'ssid': 'SSID', 'pwd': 'PWD', 'hidden': 0}
        """
        self.data = data

    @property
    def ssid(self) -> str:
        return self.data["ssid"]

    @property
    def password(self) -> str:
        return self.data["pwd"]

    @property
    def ssid_hidden(self) -> bool:
        return self.data["hidden"] == 1


class WifiRepeater(Device):
    """Device class for Xiaomi Mi WiFi Repeater 2."""

    @command(
        default_output=format_output(
            "",
            "Access policy: {result.access_policy}\n"
            "Associated stations: {result.associated_stations}\n",
        )
    )
    def status(self) -> WifiRepeaterStatus:
        """Return the associated stations."""
        return WifiRepeaterStatus(self.send("miIO.get_repeater_sta_info"))

    @command(
        default_output=format_output(
            "",
            "SSID: {result.ssid}\n"
            "Password: {result.password}\n"
            "SSID hidden: {result.ssid_hidden}\n",
        )
    )
    def configuration(self) -> WifiRepeaterConfiguration:
        """Return the configuration of the accesspoint."""
        return WifiRepeaterConfiguration(self.send("miIO.get_repeater_ap_info"))

    @command(
        click.argument("wifi_roaming", type=bool),
        default_output=format_output(
            lambda led: "Turning on WiFi roaming" if led else "Turning off WiFi roaming"
        ),
    )
    def set_wifi_roaming(self, wifi_roaming: bool):
        """Turn the WiFi roaming on/off."""
        return self.send(
            "miIO.switch_wifi_explorer", [{"wifi_explorer": int(wifi_roaming)}]
        )

    @command(
        click.argument("ssid", type=str),
        click.argument("password", type=str),
        click.argument("ssid_hidden", type=bool),
        default_output=format_output("Setting accesspoint configuration"),
    )
    def set_configuration(self, ssid: str, password: str, ssid_hidden: bool = False):
        """Update the configuration of the accesspoint."""
        return self.send(
            "miIO.switch_wifi_ssid",
            [
                {
                    "ssid": ssid,
                    "pwd": password,
                    "hidden": int(ssid_hidden),
                    "wifi_explorer": 0,
                }
            ],
        )

    @command(
        default_output=format_output(
            lambda result: "WiFi roaming is enabled"
            if result
            else "WiFi roaming is disabled"
        )
    )
    def wifi_roaming(self) -> bool:
        """Return the roaming setting."""
        return self.info().raw["desc"]["wifi_explorer"] == 1

    @command(default_output=format_output("RSSI of the accesspoint: {result}"))
    def rssi_accesspoint(self) -> int:
        """Received signal strength indicator of the accesspoint."""
        return self.info().accesspoint["rssi"]



import enum
import logging
from datetime import timedelta
from typing import Any, Dict

import click

from .click_common import EnumType, command, format_output
from .exceptions import DeviceException
from .miot_device import DeviceStatus, MiotDevice

_LOGGER = logging.getLogger(__name__)
_MAPPING = {
    # Source http://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:air-conditioner:0000A004:xiaomi-mc4:1
    # Air Conditioner (siid=2)
    "power": {"siid": 2, "piid": 1},
    "mode": {"siid": 2, "piid": 2},
    "target_temperature": {"siid": 2, "piid": 4},
    "eco": {"siid": 2, "piid": 7},
    "heater": {"siid": 2, "piid": 9},
    "dryer": {"siid": 2, "piid": 10},
    "sleep_mode": {"siid": 2, "piid": 11},
    # Fan Control (siid=3)
    "fan_speed": {"siid": 3, "piid": 2},
    "vertical_swing": {"siid": 3, "piid": 4},
    # Environment (siid=4)
    "temperature": {"siid": 4, "piid": 7},
    # Alarm (siid=5)
    "buzzer": {"siid": 5, "piid": 1},
    # Indicator Light (siid=6)
    "led": {"siid": 6, "piid": 1},
    # Electricity (siid=8)
    "electricity": {"siid": 8, "piid": 1},
    # Maintenance (siid=9)
    "clean": {"siid": 9, "piid": 1},
    "running_duration": {"siid": 9, "piid": 5},
    # Enhance (siid=10)
    "fan_speed_percent": {"siid": 10, "piid": 1},
    "timer": {"siid": 10, "piid": 3},
}

CLEANING_STAGES = [
    "Stopped",
    "Condensing water",
    "Frosting the surface",
    "Defrosting the surface",
    "Drying",
]


class AirConditionerMiotException(DeviceException):
    pass


class CleaningStatus(DeviceStatus):
    def __init__(self, status: str):
        """Auto clean mode indicator.

        Value format: <int>,<int>,<int>,<int>
        Integer 1: whether auto cleaning mode started.
        Integer 2: current progress in percent.
        Integer 3: which stage it is currently under (see CLEANING_STAGE list).
        Integer 4: if current operation could be cancelled.

        Example auto clean indicator 1: 0,100,0,1
        indicates the auto clean mode has finished or not started yet.
        Example auto clean indicator 2: 1,22,1,1
        indicates auto clean mode finished 22%, it is condensing water and can be cancelled.
        Example auto clean indicator 3: 1,72,4,0
        indicates auto clean mode finished 72%, it is drying and cannot be cancelled.

        Only write 1 or 0 to it would start or abort the auto clean mode.
        """
        self.status = [int(value) for value in status.split(",")]

    @property
    def cleaning(self) -> bool:
        return bool(self.status[0])

    @property
    def progress(self) -> int:
        return int(self.status[1])

    @property
    def stage(self) -> str:
        try:
            return CLEANING_STAGES[self.status[2]]
        except KeyError:
            return "Unknown stage"

    @property
    def cancellable(self) -> bool:
        return bool(self.status[3])


class OperationMode(enum.Enum):
    Cool = 2
    Dry = 3
    Fan = 4
    Heat = 5


class FanSpeed(enum.Enum):
    Auto = 0
    Level1 = 1
    Level2 = 2
    Level3 = 3
    Level4 = 4
    Level5 = 5
    Level6 = 6
    Level7 = 7


class TimerStatus(DeviceStatus):
    def __init__(self, status):
        """Countdown timer indicator.

        Value format: <int>,<int>,<int>,<int>
        Integer 1: whether the timer is enabled.
        Integer 2: countdown timer setting value in minutes.
        Integer 3: the device would be powered on (1) or powered off (0) after timeout.
        Integer 4: the remaining countdown time in minutes.

        Example timer value 1: 1,120,0,103
        indicates the device would be turned off after 120 minutes, remaining 103 minutes.
        Example timer value 2: 1,60,1,60
        indicates the device would be turned on after 60 minutes, remaining 60 minutes.
        Example timer value 3: 0,0,0,0
        indicates countdown timer not set.

        Write the first three integers would set the correct countdown timer.
        Also, if the countdown minutes set to 0, the timer would be disabled.
        """
        self.status = [int(value) for value in status.split(",")]

    @property
    def enabled(self) -> bool:
        return bool(self.status[0])

    @property
    def countdown(self) -> timedelta:
        return timedelta(minutes=self.status[1])

    @property
    def power_on(self) -> bool:
        return bool(self.status[2])

    @property
    def time_left(self) -> timedelta:
        return timedelta(minutes=self.status[3])


class AirConditionerMiotStatus(DeviceStatus):
    """Container for status reports from the air conditioner (MIoT)."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """
        Response (MIoT format) of a Mi Smart Air Conditioner A (xiaomi.aircondition.mc4)
        [
            {'did': 'power', 'siid': 2, 'piid': 1, 'code': 0, 'value': False},
            {'did': 'mode', 'siid': 2, 'piid': 2, 'code': 0, 'value': 2},
            {'did': 'target_temperature', 'siid': 2, 'piid': 4, 'code': 0, 'value': 26.5},
            {'did': 'eco', 'siid': 2, 'piid': 7, 'code': 0, 'value': False},
            {'did': 'heater', 'siid': 2, 'piid': 9, 'code': 0, 'value': True},
            {'did': 'dryer', 'siid': 2, 'piid': 10, 'code': 0, 'value': True},
            {'did': 'sleep_mode', 'siid': 2, 'piid': 11, 'code': 0, 'value': False},
            {'did': 'fan_speed', 'siid': 3, 'piid': 2, 'code': 0, 'value': 0},
            {'did': 'vertical_swing', 'siid': 3, 'piid': 4, 'code': 0, 'value': True},
            {'did': 'temperature', 'siid': 4, 'piid': 7, 'code': 0, 'value': 28.4},
            {'did': 'buzzer', 'siid': 5, 'piid': 1, 'code': 0, 'value': False},
            {'did': 'led', 'siid': 6, 'piid': 1, 'code': 0, 'value': False},
            {'did': 'electricity', 'siid': 8, 'piid': 1, 'code': 0, 'value': 0.0},
            {'did': 'clean', 'siid': 9, 'piid': 1, 'code': 0, 'value': '0,100,1,1'},
            {'did': 'running_duration', 'siid': 9, 'piid': 5, 'code': 0, 'value': 151.0},
            {'did': 'fan_speed_percent', 'siid': 10, 'piid': 1, 'code': 0, 'value': 101},
            {'did': 'timer', 'siid': 10, 'piid': 3, 'code': 0, 'value': '0,0,0,0'}
        ]

        """
        self.data = data

    @property
    def is_on(self) -> bool:
        """True if the device is turned on."""
        return self.data["power"]

    @property
    def power(self) -> str:
        """Current power state."""
        return "on" if self.is_on else "off"

    @property
    def mode(self) -> OperationMode:
        """Current operation mode."""
        return OperationMode(self.data["mode"])

    @property
    def target_temperature(self) -> float:
        """Target temperature in Celsius."""
        return self.data["target_temperature"]

    @property
    def eco(self) -> bool:
        """True if ECO mode is on."""
        return self.data["eco"]

    @property
    def heater(self) -> bool:
        """True if aux heat mode is on."""
        return self.data["heater"]

    @property
    def dryer(self) -> bool:
        """True if aux dryer mode is on."""
        return self.data["dryer"]

    @property
    def sleep_mode(self) -> bool:
        """True if sleep mode is on."""
        return self.data["sleep_mode"]

    @property
    def fan_speed(self) -> FanSpeed:
        """Current Fan speed."""
        return FanSpeed(self.data["fan_speed"])

    @property
    def vertical_swing(self) -> bool:
        """True if vertical swing is on."""
        return self.data["vertical_swing"]

    @property
    def temperature(self) -> float:
        """Current ambient temperature in Celsius."""
        return self.data["temperature"]

    @property
    def buzzer(self) -> bool:
        """True if buzzer is on."""
        return self.data["buzzer"]

    @property
    def led(self) -> bool:
        """True if LED is on."""
        return self.data["led"]

    @property
    def electricity(self) -> float:
        """Power consumption accumulation in kWh."""
        return self.data["electricity"]

    @property
    def clean(self) -> CleaningStatus:
        """Auto clean mode indicator."""
        return CleaningStatus(self.data["clean"])

    @property
    def total_running_duration(self) -> timedelta:
        """Total running duration in hours."""
        return timedelta(hours=self.data["running_duration"])

    @property
    def fan_speed_percent(self) -> int:
        """Current fan speed in percent."""
        return self.data["fan_speed_percent"]

    @property
    def timer(self) -> TimerStatus:
        """Countdown timer indicator."""
        return TimerStatus(self.data["timer"])


class AirConditionerMiot(MiotDevice):
    """Main class representing the air conditioner which uses MIoT protocol."""

    mapping = _MAPPING

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "Target Temperature: {result.target_temperature} ℃\n"
            "ECO Mode: {result.eco}\n"
            "Heater: {result.heater}\n"
            "Dryer: {result.dryer}\n"
            "Sleep Mode: {result.sleep_mode}\n"
            "Fan Speed: {result.fan_speed}\n"
            "Vertical Swing: {result.vertical_swing}\n"
            "Room Temperature: {result.temperature} ℃\n"
            "Buzzer: {result.buzzer}\n"
            "LED: {result.led}\n"
            "Electricity: {result.electricity}kWh\n"
            "Clean: {result.clean}\n"
            "Running Duration: {result.total_running_duration}\n"
            "Fan percent: {result.fan_speed_percent}\n"
            "Timer: {result.timer}\n",
        )
    )
    def status(self) -> AirConditionerMiotStatus:
        """Retrieve properties."""

        return AirConditionerMiotStatus(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping()
            }
        )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.set_property("power", True)

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.set_property("power", False)

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting operation mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set operation mode."""
        return self.set_property("mode", mode.value)

    @command(
        click.argument("target_temperature", type=float),
        default_output=format_output(
            "Setting target temperature to {target_temperature}"
        ),
    )
    def set_target_temperature(self, target_temperature: float):
        """Set target temperature in Celsius."""
        if (
            target_temperature < 16.0
            or target_temperature > 31.0
            or target_temperature % 0.5 != 0
        ):
            raise AirConditionerMiotException(
                "Invalid target temperature: %s" % target_temperature
            )
        return self.set_property("target_temperature", target_temperature)

    @command(
        click.argument("eco", type=bool),
        default_output=format_output(
            lambda eco: "Turning on ECO mode" if eco else "Turning off ECO mode"
        ),
    )
    def set_eco(self, eco: bool):
        """Turn ECO mode on/off."""
        return self.set_property("eco", eco)

    @command(
        click.argument("heater", type=bool),
        default_output=format_output(
            lambda heater: "Turning on heater" if heater else "Turning off heater"
        ),
    )
    def set_heater(self, heater: bool):
        """Turn aux heater mode on/off."""
        return self.set_property("heater", heater)

    @command(
        click.argument("dryer", type=bool),
        default_output=format_output(
            lambda dryer: "Turning on dryer" if dryer else "Turning off dryer"
        ),
    )
    def set_dryer(self, dryer: bool):
        """Turn aux dryer mode on/off."""
        return self.set_property("dryer", dryer)

    @command(
        click.argument("sleep_mode", type=bool),
        default_output=format_output(
            lambda sleep_mode: "Turning on sleep mode"
            if sleep_mode
            else "Turning off sleep mode"
        ),
    )
    def set_sleep_mode(self, sleep_mode: bool):
        """Turn sleep mode on/off."""
        return self.set_property("sleep_mode", sleep_mode)

    @command(
        click.argument("fan_speed", type=EnumType(FanSpeed)),
        default_output=format_output("Setting fan speed to {fan_speed}"),
    )
    def set_fan_speed(self, fan_speed: FanSpeed):
        """Set fan speed."""
        return self.set_property("fan_speed", fan_speed.value)

    @command(
        click.argument("vertical_swing", type=bool),
        default_output=format_output(
            lambda vertical_swing: "Turning on vertical swing"
            if vertical_swing
            else "Turning off vertical swing"
        ),
    )
    def set_vertical_swing(self, vertical_swing: bool):
        """Turn vertical swing on/off."""
        return self.set_property("vertical_swing", vertical_swing)

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        return self.set_property("led", led)

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        return self.set_property("buzzer", buzzer)

    @command(
        click.argument("percent", type=int),
        default_output=format_output("Setting fan percent to {percent}%"),
    )
    def set_fan_speed_percent(self, fan_speed_percent):
        """Set fan speed in percent, should be  between 1 to 100 or 101(auto)."""
        if fan_speed_percent < 1 or fan_speed_percent > 101:
            raise AirConditionerMiotException(
                "Invalid fan percent: %s" % fan_speed_percent
            )
        return self.set_property("fan_speed_percent", fan_speed_percent)

    @command(
        click.argument("minutes", type=int),
        click.argument("delay_on", type=bool),
        default_output=format_output(
            lambda minutes, delay_on: "Setting timer to delay on after "
            + str(minutes)
            + " minutes"
            if delay_on
            else "Setting timer to delay off after " + str(minutes) + " minutes"
        ),
    )
    def set_timer(self, minutes, delay_on):
        """Set countdown timer minutes and if it would be turned on after timeout.

        Set minutes to 0 would disable the timer.
        """
        return self.set_property(
            "timer", ",".join(["1", str(minutes), str(int(delay_on))])
        )

    @command(
        click.argument("clean", type=bool),
        default_output=format_output(
            lambda clean: "Begin auto cleanning" if clean else "Abort auto cleaning"
        ),
    )
    def set_clean(self, clean):
        """Start or abort clean mode."""
        return self.set_property("clean", str(int(clean)))



import logging
from enum import Enum
from functools import partial
from typing import Any, Dict, Union

import click

from .click_common import EnumType, LiteralParamType, command
from .device import Device, DeviceStatus  # noqa: F401
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)


# partial is required here for str2bool, see https://stackoverflow.com/a/40339397
class MiotValueType(Enum):
    def _str2bool(x):
        """Helper to convert string to boolean."""
        return x.lower() in ("true", "1")

    Int = int
    Float = float
    Bool = partial(_str2bool)
    Str = str


MiotMapping = Dict[str, Dict[str, Any]]


class MiotDevice(Device):
    """Main class representing a MIoT device."""

    mapping: MiotMapping

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        timeout: int = None,
        *,
        mapping: MiotMapping = None,
    ):
        """Overloaded to accept keyword-only `mapping` parameter."""
        super().__init__(ip, token, start_id, debug, lazy_discover, timeout)

        if mapping is None and not hasattr(self, "mapping"):
            raise DeviceException(
                "Neither the class nor the parameter defines the mapping"
            )

        if mapping is not None:
            self.mapping = mapping

    def get_properties_for_mapping(self, *, max_properties=15) -> list:
        """Retrieve raw properties based on mapping."""

        # We send property key in "did" because it's sent back via response and we can identify the property.
        properties = [
            {"did": k, **v} for k, v in self.mapping.items() if "aiid" not in v
        ]

        return self.get_properties(
            properties, property_getter="get_properties", max_properties=max_properties
        )

    @command(
        click.argument("name", type=str),
        click.argument("params", type=LiteralParamType(), required=False),
    )
    def call_action(self, name: str, params=None):
        """Call an action by a name in the mapping."""
        if name not in self.mapping:
            raise DeviceException(f"Unable to find {name} in the mapping")

        action = self.mapping[name]

        if "siid" not in action or "aiid" not in action:
            raise DeviceException(f"{name} is not an action (missing siid or aiid)")

        return self.call_action_by(action["siid"], action["aiid"], params)

    @command(
        click.argument("siid", type=int),
        click.argument("aiid", type=int),
        click.argument("params", type=LiteralParamType(), required=False),
    )
    def call_action_by(self, siid, aiid, params=None):
        """Call an action."""
        if params is None:
            params = []
        payload = {
            "did": f"call-{siid}-{aiid}",
            "siid": siid,
            "aiid": aiid,
            "in": params,
        }

        return self.send("action", payload)

    @command(
        click.argument("siid", type=int),
        click.argument("piid", type=int),
    )
    def get_property_by(self, siid: int, piid: int):
        """Get a single property (siid/piid)."""
        return self.send(
            "get_properties", [{"did": f"{siid}-{piid}", "siid": siid, "piid": piid}]
        )

    @command(
        click.argument("siid", type=int),
        click.argument("piid", type=int),
        click.argument("value"),
        click.argument(
            "value_type", type=EnumType(MiotValueType), required=False, default=None
        ),
    )
    def set_property_by(
        self,
        siid: int,
        piid: int,
        value: Union[int, float, str, bool],
        value_type: Any = None,
    ):
        """Set a single property (siid/piid) to given value.

        value_type can be given to convert the value to wanted type, allowed types are:
        int, float, bool, str
        """
        if value_type is not None:
            value = value_type.value(value)

        return self.send(
            "set_properties",
            [{"did": f"set-{siid}-{piid}", "siid": siid, "piid": piid, "value": value}],
        )

    def set_property(self, property_key: str, value):
        """Sets property value using the existing mapping."""
        return self.send(
            "set_properties",
            [{"did": property_key, **self.mapping[property_key], "value": value}],
        )



from miio import DeviceStatus


def test_multiple():
    class MultipleProperties(DeviceStatus):
        @property
        def first(self):
            return "first"

        @property
        def second(self):
            return "second"

    assert (
        repr(MultipleProperties()) == "<MultipleProperties first=first second=second>"
    )


def test_empty():
    class EmptyStatus(DeviceStatus):
        pass

    assert repr(EmptyStatus() == "<EmptyStatus>")


def test_exception():
    class StatusWithException(DeviceStatus):
        @property
        def raise_exception(self):
            raise Exception("test")

    assert (
        repr(StatusWithException()) == "<StatusWithException raise_exception=Exception>"
    )


def test_inheritance():
    class Parent(DeviceStatus):
        @property
        def from_parent(self):
            return True

    class Child(Parent):
        @property
        def from_child(self):
            return True

    assert repr(Child()) == "<Child from_child=True from_parent=True>"


def test_list():
    class List(DeviceStatus):
        @property
        def return_list(self):
            return [0, 1, 2]

    assert repr(List()) == "<List return_list=[0, 1, 2]>"


def test_none():
    class NoneStatus(DeviceStatus):
        @property
        def return_none(self):
            return None

    assert repr(NoneStatus()) == "<NoneStatus return_none=None>"



import enum
import logging
from typing import Any, Dict

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_FAN_LESHOW_SS4 = "leshow.fan.ss4"

AVAILABLE_PROPERTIES_COMMON = [
    "power",
    "mode",
    "blow",
    "timer",
    "sound",
    "yaw",
    "fault",
]

AVAILABLE_PROPERTIES = {
    MODEL_FAN_LESHOW_SS4: AVAILABLE_PROPERTIES_COMMON,
}


class FanLeshowException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Manual = 0
    Sleep = 1
    Strong = 2
    Natural = 3


class FanLeshowStatus(DeviceStatus):
    """Container for status reports from the Xiaomi Rosou SS4 Ventilator."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Response of a Leshow Fan SS4 (leshow.fan.ss4):

        {'power': 1, 'mode': 2, 'blow': 100, 'timer': 0,
         'sound': 1, 'yaw': 0, 'fault': 0}
        """
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return "on" if self.data["power"] == 1 else "off"

    @property
    def is_on(self) -> bool:
        """True if device is turned on."""
        return self.data["power"] == 1

    @property
    def mode(self) -> OperationMode:
        """Operation mode."""
        return OperationMode(self.data["mode"])

    @property
    def speed(self) -> int:
        """Speed of the fan in percent."""
        return self.data["blow"]

    @property
    def buzzer(self) -> bool:
        """True if buzzer is turned on."""
        return self.data["sound"] == 1

    @property
    def oscillate(self) -> bool:
        """True if oscillation is enabled."""
        return self.data["yaw"] == 1

    @property
    def delay_off_countdown(self) -> int:
        """Countdown until turning off in minutes."""
        return self.data["timer"]

    @property
    def error_detected(self) -> bool:
        """True if a fault was detected."""
        return self.data["fault"] == 1


class FanLeshow(Device):
    """Main class representing the Xiaomi Rosou SS4 Ventilator."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_FAN_LESHOW_SS4,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_FAN_LESHOW_SS4

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "Speed: {result.speed}\n"
            "Buzzer: {result.buzzer}\n"
            "Oscillate: {result.oscillate}\n"
            "Power-off time: {result.delay_off_countdown}\n"
            "Error detected: {result.error_detected}\n",
        )
    )
    def status(self) -> FanLeshowStatus:
        """Retrieve properties."""
        properties = AVAILABLE_PROPERTIES[self.model]
        values = self.get_properties(properties, max_properties=15)

        return FanLeshowStatus(dict(zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", [1])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", [0])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode (manual, natural, sleep, strong)."""
        return self.send("set_mode", [mode.value])

    @command(
        click.argument("speed", type=int),
        default_output=format_output("Setting speed of the manual mode to {speed}"),
    )
    def set_speed(self, speed: int):
        """Set a speed level between 0 and 100."""
        if speed < 0 or speed > 100:
            raise FanLeshowException("Invalid speed: %s" % speed)

        return self.send("set_blow", [speed])

    @command(
        click.argument("oscillate", type=bool),
        default_output=format_output(
            lambda oscillate: "Turning on oscillate"
            if oscillate
            else "Turning off oscillate"
        ),
    )
    def set_oscillate(self, oscillate: bool):
        """Set oscillate on/off."""
        return self.send("set_yaw", [int(oscillate)])

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        return self.send("set_sound", [int(buzzer)])

    @command(
        click.argument("minutes", type=int),
        default_output=format_output("Setting delayed turn off to {minutes} minutes"),
    )
    def delay_off(self, minutes: int):
        """Set delay off minutes."""

        if minutes < 0 or minutes > 540:
            raise FanLeshowException(
                "Invalid value for a delayed turn off: %s" % minutes
            )

        return self.send("set_timer", [minutes])



from unittest import TestCase

import pytest

from miio import PhilipsMoonlight
from miio.philips_moonlight import PhilipsMoonlightException, PhilipsMoonlightStatus
from miio.utils import int_to_rgb, rgb_to_int

from .dummies import DummyDevice


class DummyPhilipsMoonlight(DummyDevice, PhilipsMoonlight):
    def __init__(self, *args, **kwargs):
        self.state = {
            "pow": "on",
            "sta": 0,
            "bri": 1,
            "rgb": 16741971,
            "cct": 1,
            "snm": 0,
            "spr": 0,
            "spt": 15,
            "wke": 0,
            "bl": 1,
            "ms": 1,
            "mb": 1,
            "wkp": [0, 24, 0],
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("pow", x),
            "set_bright": lambda x: self._set_state("bri", x),
            "set_cct": lambda x: self._set_state("cct", x),
            "set_rgb": lambda x: self._set_state("rgb", [rgb_to_int(x)]),
            "apply_fixed_scene": lambda x: self._set_state("snm", x),
            "go_night": lambda x: self._set_state("snm", [6]),
            "set_bricct": lambda x: (
                self._set_state("bri", [x[0]]),
                self._set_state("cct", [x[1]]),
            ),
            "set_brirgb": lambda x: (
                self._set_state("rgb", [rgb_to_int((x[0], x[1], x[2]))]),
                self._set_state("bri", [x[3]]),
            ),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def philips_moonlight(request):
    request.cls.device = DummyPhilipsMoonlight()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("philips_moonlight")
class TestPhilipsMoonlight(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            PhilipsMoonlightStatus(self.device.start_state)
        )

        assert self.is_on() is True
        assert self.state().brightness == self.device.start_state["bri"]
        assert self.state().color_temperature == self.device.start_state["cct"]
        assert self.state().rgb == int_to_rgb(int(self.device.start_state["rgb"]))
        assert self.state().scene == self.device.start_state["snm"]

    def test_set_brightness(self):
        def brightness():
            return self.device.status().brightness

        self.device.set_brightness(1)
        assert brightness() == 1
        self.device.set_brightness(50)
        assert brightness() == 50
        self.device.set_brightness(100)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness(-1)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness(0)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness(101)

    def test_set_rgb(self):
        def rgb():
            return self.device.status().rgb

        self.device.set_rgb((0, 0, 1))
        assert rgb() == (0, 0, 1)
        self.device.set_rgb((255, 255, 0))
        assert rgb() == (255, 255, 0)
        self.device.set_rgb((255, 255, 255))
        assert rgb() == (255, 255, 255)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_rgb((-1, 0, 0))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_rgb((256, 0, 0))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_rgb((0, -1, 0))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_rgb((0, 256, 0))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_rgb((0, 0, -1))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_rgb((0, 0, 256))

    def test_set_color_temperature(self):
        def color_temperature():
            return self.device.status().color_temperature

        self.device.set_color_temperature(20)
        assert color_temperature() == 20
        self.device.set_color_temperature(30)
        assert color_temperature() == 30
        self.device.set_color_temperature(10)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_color_temperature(-1)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_color_temperature(0)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_color_temperature(101)

    def test_set_brightness_and_color_temperature(self):
        def color_temperature():
            return self.device.status().color_temperature

        def brightness():
            return self.device.status().brightness

        self.device.set_brightness_and_color_temperature(20, 21)
        assert brightness() == 20
        assert color_temperature() == 21
        self.device.set_brightness_and_color_temperature(31, 30)
        assert brightness() == 31
        assert color_temperature() == 30
        self.device.set_brightness_and_color_temperature(10, 11)
        assert brightness() == 10
        assert color_temperature() == 11

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_color_temperature(-1, 10)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_color_temperature(10, -1)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_color_temperature(0, 10)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_color_temperature(10, 0)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_color_temperature(101, 10)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_color_temperature(10, 101)

    def test_set_brightness_and_rgb(self):
        def brightness():
            return self.device.status().brightness

        def rgb():
            return self.device.status().rgb

        self.device.set_brightness_and_rgb(20, (0, 0, 0))
        assert brightness() == 20
        assert rgb() == (0, 0, 0)
        self.device.set_brightness_and_rgb(31, (255, 0, 0))
        assert brightness() == 31
        assert rgb() == (255, 0, 0)
        self.device.set_brightness_and_rgb(100, (255, 255, 255))
        assert brightness() == 100
        assert rgb() == (255, 255, 255)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(-1, 10)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(0, 10)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(101, 10)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(10, (-1, 0, 0))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(10, (256, 0, 0))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(10, (0, -1, 0))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(10, (0, 256, 0))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(10, (0, 0, -1))

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_brightness_and_rgb(10, (0, 0, 256))

    def test_set_scene(self):
        def scene():
            return self.device.status().scene

        self.device.set_scene(1)
        assert scene() == 1
        self.device.set_scene(6)
        assert scene() == 6

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_scene(-1)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_scene(0)

        with pytest.raises(PhilipsMoonlightException):
            self.device.set_scene(7)



import enum
import logging
import warnings

import click

from .click_common import command, format_output
from .device import Device, DeviceStatus

_LOGGER = logging.getLogger(__name__)


class PlayState(enum.Enum):
    Playing = "PLAYING"
    Stopped = "STOPPED"
    Paused = "PAUSED_PLAYBACK"
    NoMedia = "NO_MEDIA_PRESENT"
    Transitioning = "TRANSITIONING"


class TransportChannel(enum.Enum):
    Playlist = "PLAYLIST"
    OneTime = "ONETIME"
    Auxiliary = "AUX"
    Bluetooth = "BT"
    Radio = "RADIO"
    Air = "AIR"
    Qplay = "QPLAY"


class WifiSpeakerStatus(DeviceStatus):
    """Container of a speaker state.

    This contains information such as the name of the device, and what is currently
    being played by it.
    """

    def __init__(self, data):
        """Example response of a xiaomi.wifispeaker.v2:

        {"DeviceName": "Mi Internet Speaker", "channel_title\": "XXX",
         "current_state": "PLAYING", "hardware_version": "S602",
         "play_mode": "REPEAT_ALL", "track_artist": "XXX",
         "track_duration": "00:04:58", "track_title": "XXX",
         "transport_channel": "PLAYLIST"}
        """
        self.data = data

    @property
    def device_name(self) -> str:
        """Name of the device."""
        return self.data["DeviceName"]

    @property
    def channel(self) -> str:
        """Name of the channel."""
        return self.data["channel_title"]

    @property
    def state(self) -> PlayState:
        """State of the device, e.g. PLAYING."""
        return PlayState(self.data["current_state"])

    @property
    def hardware_version(self) -> str:
        return self.data["hardware_version"]

    @property
    def play_mode(self):
        """Play mode such as REPEAT_ALL."""
        # note: this can be enumized when all values are known
        return self.data["play_mode"]

    @property
    def track_artist(self) -> str:
        """Artist of the current track."""
        return self.data["track_artist"]

    @property
    def track_title(self) -> str:
        """Title of the current track."""
        return self.data["track_title"]

    @property
    def track_duration(self) -> str:
        """Total duration of the current track."""
        return self.data["track_duration"]

    @property
    def transport_channel(self) -> TransportChannel:
        """Transport channel, e.g. PLAYLIST."""
        return TransportChannel(self.data["transport_channel"])


class WifiSpeaker(Device):
    """Device class for Xiaomi Smart Wifi Speaker."""

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Please help to complete this by providing more "
            "information about possible values for `state`, "
            "`play_mode` and `transport_channel`.",
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)

    @command(
        default_output=format_output(
            "",
            "Device name: {result.device_name}\n"
            "Channel: {result.channel}\n"
            "State: {result.state}\n"
            "Play mode: {result.play_mode}\n"
            "Track artist: {result.track_artist}\n"
            "Track title: {result.track_title}\n"
            "Track duration: {result.track_duration}\n"
            "Transport channel: {result.transport_channel}\n"
            "Hardware version: {result.hardware_version}\n",
        )
    )
    def status(self) -> WifiSpeakerStatus:
        """Return device status."""
        return WifiSpeakerStatus(self.send("get_prop", ["umi"]))

    @command(default_output=format_output("Powering on"))
    def power(self):
        """Toggle power on and off."""
        # is this a toggle?
        return self.send("power")

    @command(default_output=format_output("Toggling play"))
    def toggle(self):
        """Toggle play."""
        return self.send("toggle")

    @command(
        click.argument("amount", type=int),
        default_output=format_output("Increasing volume by {amount} percent"),
    )
    def volume_up(self, amount: int = 5):
        """Set volume up."""
        return self.send("vol_up", [amount])

    @command(
        click.argument("amount", type=int),
        default_output=format_output("Decreasing volume by {amount} percent"),
    )
    def volume_down(self, amount: int = 5):
        """Set volume down."""
        return self.send("vol_down", [amount])

    @command(default_output=format_output("Playing previous track"))
    def track_previous(self):
        """Move to previous track."""
        return self.send("previous_track")

    @command(default_output=format_output("Playing next track"))
    def track_next(self):
        """Move to next track."""
        return self.send("next_track")

    @command(default_output=format_output("Switching to the next transport channel"))
    def channel_next(self):
        """Change transport channel."""
        return self.send("next_channel")

    @command(default_output=format_output("Track position: {result.rel_time}"))
    def track_position(self):
        """Return current track position."""
        return self.send("get_prop", ["rel_time"])

    def volume(self):
        """Speaker volume."""
        return self.send("get_prop", ["volume"])



import enum
import logging
from typing import Optional

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_ACPARTNER_V1 = "lumi.acpartner.v1"
MODEL_ACPARTNER_V2 = "lumi.acpartner.v2"
MODEL_ACPARTNER_V3 = "lumi.acpartner.v3"

MODELS_SUPPORTED = [MODEL_ACPARTNER_V1, MODEL_ACPARTNER_V2, MODEL_ACPARTNER_V3]


class AirConditioningCompanionException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Heat = 0
    Cool = 1
    Auto = 2
    Dehumidify = 3
    Ventilate = 4


class FanSpeed(enum.Enum):
    Low = 0
    Medium = 1
    High = 2
    Auto = 3


class SwingMode(enum.Enum):
    On = "0"
    Off = "1"
    Unknown2 = "2"
    Unknown7 = "7"
    ChigoOn = "C"
    ChigoOff = "D"


class Power(enum.Enum):
    On = 1
    Off = 0


class Led(enum.Enum):
    On = "0"
    Off = "A"


STORAGE_SLOT_ID = 30
POWER_OFF = "off"

# Command templates per model number (f.e. 0180111111)
# [po], [mo], [wi], [sw], [tt], [tt1], [tt4] and [tt7] are markers which will be replaced
DEVICE_COMMAND_TEMPLATES = {
    "fallback": {"deviceType": "generic", "base": "[po][mo][wi][sw][tt][li]"},
    "0100010727": {
        "deviceType": "gree_2",
        "base": "[po][mo][wi][sw][tt]1100190[tt1]205002102000[tt7]0190[tt1]207002000000[tt4]",
        "off": "01011101004000205002112000D04000207002000000A0",
    },
    "0100004795": {
        "deviceType": "gree_8",
        "base": "[po][mo][wi][sw][tt][li]10009090000500",
    },
    "0180333331": {"deviceType": "haier_1", "base": "[po][mo][wi][sw][tt]1"},
    "0180666661": {"deviceType": "aux_1", "base": "[po][mo][wi][sw][tt]1"},
    "0180777771": {"deviceType": "chigo_1", "base": "[po][mo][wi][sw][tt]1"},
}


class AirConditioningCompanionStatus(DeviceStatus):
    """Container for status reports of the Xiaomi AC Companion."""

    def __init__(self, data):
        """Device model: lumi.acpartner.v2.

        Response of "get_model_and_state":
        ['010500978022222102', '010201190280222221', '2']

        AC turned on by set_power=on:
        ['010507950000257301', '011001160100002573', '807']

        AC turned off by set_power=off:
        ['010507950000257301', '010001160100002573', '6']
        ...
        ['010507950000257301', '010001160100002573', '1']

        Example data payload:
        { 'model_and_state': ['010500978022222102', '010201190280222221', '2'],
          'power_socket': 'on' }
        """
        self.data = data
        self.model = data["model_and_state"][0]
        self.state = data["model_and_state"][1]

    @property
    def load_power(self) -> int:
        """Current power load of the air conditioner."""
        return int(self.data["model_and_state"][2])

    @property
    def power_socket(self) -> Optional[str]:
        """Current socket power state."""
        if "power_socket" in self.data and self.data["power_socket"] is not None:
            return self.data["power_socket"]

        return None

    @property
    def air_condition_model(self) -> bytes:
        """Model of the air conditioner."""
        return bytes.fromhex(self.model)

    @property
    def model_format(self) -> int:
        """Version number of the model format."""
        return self.air_condition_model[0]

    @property
    def device_type(self) -> int:
        """Device type identifier."""
        return self.air_condition_model[1]

    @property
    def air_condition_brand(self) -> int:
        """Brand of the air conditioner.

        Known brand ids are 0x0182, 0x0097, 0x0037, 0x0202, 0x02782, 0x0197, 0x0192.
        """
        return int(self.air_condition_model[2:4].hex(), 16)

    @property
    def air_condition_remote(self) -> int:
        """Remote id.

        Known remote ids:
        * 0x80111111, 0x80111112 (brand: 0x0182)
        * 0x80222221 (brand: 0x0097)
        * 0x80333331 (brand: 0x0037)
        * 0x80444441 (brand: 0x0202)
        * 0x80555551 (brand: 0x2782)
        * 0x80777771 (brand: 0x0197)
        * 0x80666661 (brand: 0x0192)
        """
        return int(self.air_condition_model[4:8].hex(), 16)

    @property
    def state_format(self) -> int:
        """Version number of the state format.

        Known values are: 1, 2, 3
        """
        return int(self.air_condition_model[8])

    @property
    def air_condition_configuration(self) -> int:
        return self.state[2:10]

    @property
    def power(self) -> str:
        """Current power state."""
        return "on" if int(self.state[2:3]) == Power.On.value else "off"

    @property
    def led(self) -> Optional[bool]:
        """Current LED state."""
        state = self.state[8:9]
        if state == Led.On.value:
            return True

        if state == Led.Off.value:
            return False

        _LOGGER.info("Unsupported LED state: %s", state)
        return None

    @property
    def is_on(self) -> bool:
        """True if the device is turned on."""
        return self.power == "on"

    @property
    def target_temperature(self) -> Optional[int]:
        """Target temperature."""
        try:
            return int(self.state[6:8], 16)
        except TypeError:
            return None

    @property
    def swing_mode(self) -> Optional[SwingMode]:
        """Current swing mode."""
        try:
            mode = self.state[5:6]
            return SwingMode(mode)
        except TypeError:
            return None

    @property
    def fan_speed(self) -> Optional[FanSpeed]:
        """Current fan speed."""
        try:
            speed = int(self.state[4:5])
            return FanSpeed(speed)
        except TypeError:
            return None

    @property
    def mode(self) -> Optional[OperationMode]:
        """Current operation mode."""
        try:
            mode = int(self.state[3:4])
            return OperationMode(mode)
        except TypeError:
            return None


class AirConditioningCompanion(Device):
    """Main class representing Xiaomi Air Conditioning Companion V1 and V2."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_ACPARTNER_V2,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in MODELS_SUPPORTED:
            self.model = model
        else:
            self.model = MODEL_ACPARTNER_V2
            _LOGGER.error(
                "Device model %s unsupported. Falling back to %s.", model, self.model
            )

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Load power: {result.load_power}\n"
            "Air Condition model: {result.air_condition_model}\n"
            "LED: {result.led}\n"
            "Target temperature: {result.target_temperature} °C\n"
            "Swing mode: {result.swing_mode}\n"
            "Fan speed: {result.fan_speed}\n"
            "Mode: {result.mode}\n",
        )
    )
    def status(self) -> AirConditioningCompanionStatus:
        """Return device status."""
        status = self.send("get_model_and_state")
        return AirConditioningCompanionStatus(dict(model_and_state=status))

    @command(default_output=format_output("Powering the air condition on"))
    def on(self):
        """Turn the air condition on by infrared."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering the air condition off"))
    def off(self):
        """Turn the air condition off by infrared."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("slot", type=int),
        default_output=format_output(
            "Learning infrared command into storage slot {slot}"
        ),
    )
    def learn(self, slot: int = STORAGE_SLOT_ID):
        """Learn an infrared command."""
        return self.send("start_ir_learn", [slot])

    @command(default_output=format_output("Reading learned infrared commands"))
    def learn_result(self):
        """Read the learned command."""
        return self.send("get_ir_learn_result")

    @command(
        click.argument("slot", type=int),
        default_output=format_output(
            "Learning infrared command into storage slot {slot} stopped"
        ),
    )
    def learn_stop(self, slot: int = STORAGE_SLOT_ID):
        """Stop learning of a infrared command."""
        return self.send("end_ir_learn", [slot])

    @command(
        click.argument("model", type=str),
        click.argument("code", type=str),
        default_output=format_output("Sending the supplied infrared command"),
    )
    def send_ir_code(self, model: str, code: str, slot: int = 0):
        """Play a captured command.

        :param str model: Air condition model
        :param str code: Command to execute
        :param int slot: Unknown internal register or slot
        """
        try:
            model_bytes = bytes.fromhex(model)
        except ValueError:
            raise AirConditioningCompanionException(
                "Invalid model. A hexadecimal string must be provided"
            )

        try:
            code_bytes = bytes.fromhex(code)
        except ValueError:
            raise AirConditioningCompanionException(
                "Invalid code. A hexadecimal string must be provided"
            )

        if slot < 0 or slot > 134:
            raise AirConditioningCompanionException("Invalid slot: %s" % slot)

        slot_bytes = bytes([121 + slot])

        # FE + 0487 + 00007145 + 9470 + 1FFF + 7F + FF + 06 + 0042 + 27 + 4E + 0025002D008500AC01...
        command_bytes = (
            code_bytes[0:1]
            + model_bytes[2:8]
            + b"\x94\x70\x1F\xFF"
            + slot_bytes
            + b"\xFF"
            + code_bytes[13:16]
            + b"\x27"
        )

        checksum = sum(command_bytes) & 0xFF
        command_bytes = command_bytes + bytes([checksum]) + code_bytes[18:]

        return self.send("send_ir_code", [command_bytes.hex().upper()])

    @command(
        click.argument("command", type=str),
        default_output=format_output("Sending a command to the air conditioner"),
    )
    def send_command(self, command: str):
        """Send a command to the air conditioner.

        :param str command: Command to execute
        """
        return self.send("send_cmd", [str(command)])

    @command(
        click.argument("model", type=str),
        click.argument("power", type=EnumType(Power)),
        click.argument("operation_mode", type=EnumType(OperationMode)),
        click.argument("target_temperature", type=int),
        click.argument("fan_speed", type=EnumType(FanSpeed)),
        click.argument("swing_mode", type=EnumType(SwingMode)),
        click.argument("led", type=EnumType(Led)),
        default_output=format_output("Sending a configuration to the air conditioner"),
    )
    def send_configuration(
        self,
        model: str,
        power: Power,
        operation_mode: OperationMode,
        target_temperature: int,
        fan_speed: FanSpeed,
        swing_mode: SwingMode,
        led: Led,
    ):

        prefix = str(model[0:2] + model[8:16])
        suffix = model[-1:]

        # Static turn off command available?
        if (
            (power is Power.Off)
            and (prefix in DEVICE_COMMAND_TEMPLATES)
            and (POWER_OFF in DEVICE_COMMAND_TEMPLATES[prefix])
        ):
            return self.send_command(
                prefix + DEVICE_COMMAND_TEMPLATES[prefix][POWER_OFF]
            )

        if prefix in DEVICE_COMMAND_TEMPLATES:
            configuration = prefix + DEVICE_COMMAND_TEMPLATES[prefix]["base"]
        else:
            configuration = prefix + DEVICE_COMMAND_TEMPLATES["fallback"]["base"]

        configuration = configuration.replace("[po]", str(power.value))
        configuration = configuration.replace("[mo]", str(operation_mode.value))
        configuration = configuration.replace("[wi]", str(fan_speed.value))
        configuration = configuration.replace("[sw]", str(swing_mode.value))
        configuration = configuration.replace("[tt]", format(target_temperature, "X"))
        configuration = configuration.replace("[li]", str(led.value))

        temperature = format((1 + target_temperature - 17) % 16, "X")
        configuration = configuration.replace("[tt1]", temperature)

        temperature = format((4 + target_temperature - 17) % 16, "X")
        configuration = configuration.replace("[tt4]", temperature)

        temperature = format((7 + target_temperature - 17) % 16, "X")
        configuration = configuration.replace("[tt7]", temperature)

        configuration = configuration + suffix

        return self.send_command(configuration)


class AirConditioningCompanionV3(AirConditioningCompanion):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
    ) -> None:
        super().__init__(
            ip, token, start_id, debug, lazy_discover, model=MODEL_ACPARTNER_V3
        )

    @command(default_output=format_output("Powering socket on"))
    def socket_on(self):
        """Socket power on."""
        return self.send("toggle_plug", ["on"])

    @command(default_output=format_output("Powering socket off"))
    def socket_off(self):
        """Socket power off."""
        return self.send("toggle_plug", ["off"])

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Power socket: {result.power_socket}\n"
            "Load power: {result.load_power}\n"
            "Air Condition model: {result.air_condition_model}\n"
            "LED: {result.led}\n"
            "Target temperature: {result.target_temperature} °C\n"
            "Swing mode: {result.swing_mode}\n"
            "Fan speed: {result.fan_speed}\n"
            "Mode: {result.mode}\n",
        )
    )
    def status(self) -> AirConditioningCompanionStatus:
        """Return device status."""
        status = self.send("get_model_and_state")
        power_socket = self.send("get_device_prop", ["lumi.0", "plug_state"])
        return AirConditioningCompanionStatus(
            dict(model_and_state=status, power_socket=power_socket[0])
        )



"""Xiaomi Zigbee switches."""

from enum import IntEnum

import click

from ...click_common import command
from .subdevice import SubDevice


class Switch(SubDevice):
    """Base class for one channel switch subdevice that supports on/off."""

    class ChannelMap(IntEnum):
        """Option to select wich channel to control."""

        channel_0 = 0
        channel_1 = 1
        channel_2 = 2

    @command(click.argument("channel", type=int))
    def toggle(self, channel: int = 0):
        """Toggle a channel of the switch, default channel_0."""
        return self.send_arg(
            self.setter, [self.ChannelMap(channel).name, "toggle"]
        ).pop()

    @command(click.argument("channel", type=int))
    def on(self, channel: int = 0):
        """Turn on a channel of the switch, default channel_0."""
        return self.send_arg(self.setter, [self.ChannelMap(channel).name, "on"]).pop()

    @command(click.argument("channel", type=int))
    def off(self, channel: int = 0):
        """Turn off a channel of the switch, default channel_0."""
        return self.send_arg(self.setter, [self.ChannelMap(channel).name, "off"]).pop()



from unittest import TestCase

import pytest

from miio import AirQualityMonitor
from miio.airqualitymonitor import (
    MODEL_AIRQUALITYMONITOR_B1,
    MODEL_AIRQUALITYMONITOR_S1,
    MODEL_AIRQUALITYMONITOR_V1,
    AirQualityMonitorStatus,
)

from .dummies import DummyDevice


class DummyAirQualityMonitorV1(DummyDevice, AirQualityMonitor):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_AIRQUALITYMONITOR_V1
        self.state = {
            "power": "on",
            "aqi": 34,
            "battery": 100,
            "usb_state": "off",
            "time_state": "on",
            "night_state": "on",
            "night_beg_time": "format unknown",
            "night_end_time": "format unknown",
            "sensor_state": "format unknown",
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_time_state": lambda x: self._set_state("time_state", x),
            "set_night_state": lambda x: self._set_state("night_state", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def airqualitymonitorv1(request):
    request.cls.device = DummyAirQualityMonitorV1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airqualitymonitorv1")
class TestAirQualityMonitorV1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            AirQualityMonitorStatus(self.device.start_state)
        )

        assert self.is_on() is True
        assert self.state().aqi == self.device.start_state["aqi"]
        assert self.state().battery == self.device.start_state["battery"]
        assert self.state().usb_power is (self.device.start_state["usb_state"] == "on")
        assert self.state().display_clock is (
            self.device.start_state["time_state"] == "on"
        )
        assert self.state().night_mode is (
            self.device.start_state["night_state"] == "on"
        )


class DummyAirQualityMonitorS1(DummyDevice, AirQualityMonitor):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_AIRQUALITYMONITOR_S1
        self.state = {
            "battery": 100,
            "co2": 695,
            "humidity": 62.1,
            "pm25": 19.4,
            "temperature": 27.4,
            "tvoc": 254,
        }
        self.return_values = {"get_prop": self._get_state}
        super().__init__(args, kwargs)

    def _get_state(self, props):
        """Return wanted properties."""
        return self.state


@pytest.fixture(scope="class")
def airqualitymonitors1(request):
    request.cls.device = DummyAirQualityMonitorS1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airqualitymonitors1")
class TestAirQualityMonitorS1(TestCase):
    def state(self):
        return self.device.status()

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            AirQualityMonitorStatus(self.device.start_state)
        )

        assert self.state().battery == self.device.start_state["battery"]
        assert self.state().co2 == self.device.start_state["co2"]
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().pm25 == self.device.start_state["pm25"]
        assert self.state().temperature == self.device.start_state["temperature"]
        assert self.state().tvoc == self.device.start_state["tvoc"]
        assert self.state().aqi is None
        assert self.state().usb_power is None
        assert self.state().display_clock is None
        assert self.state().night_mode is None


class DummyAirQualityMonitorB1(DummyDevice, AirQualityMonitor):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_AIRQUALITYMONITOR_B1
        self.state = {
            "co2e": 1466,
            "humidity": 59.79999923706055,
            "pm25": 2,
            "temperature": 19.799999237060547,
            "temperature_unit": "c",
            "tvoc": 1.3948699235916138,
            "tvoc_unit": "mg_m3",
        }
        self.return_values = {"get_air_data": self._get_state}
        super().__init__(args, kwargs)

    def _get_state(self, props):
        """Return wanted properties."""
        return self.state


@pytest.fixture(scope="class")
def airqualitymonitorb1(request):
    request.cls.device = DummyAirQualityMonitorB1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airqualitymonitorb1")
class TestAirQualityMonitorB1(TestCase):
    def state(self):
        return self.device.status()

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            AirQualityMonitorStatus(self.device.start_state)
        )

        assert self.state().power is None
        assert self.state().usb_power is None
        assert self.state().battery is None
        assert self.state().aqi is None
        assert self.state().temperature == self.device.start_state["temperature"]
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().co2 is None
        assert self.state().co2e == self.device.start_state["co2e"]
        assert self.state().pm25 == self.device.start_state["pm25"]
        assert self.state().tvoc == self.device.start_state["tvoc"]
        assert self.state().display_clock is None
        assert self.state().night_mode is None



import enum
import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_HUMIDIFIER_MJJSQ = "deerma.humidifier.mjjsq"
MODEL_HUMIDIFIER_JSQ = "deerma.humidifier.jsq"
MODEL_HUMIDIFIER_JSQ1 = "deerma.humidifier.jsq1"

MODEL_HUMIDIFIER_JSQ_COMMON = [
    "OnOff_State",
    "TemperatureValue",
    "Humidity_Value",
    "HumiSet_Value",
    "Humidifier_Gear",
    "Led_State",
    "TipSound_State",
    "waterstatus",
    "watertankstatus",
]

AVAILABLE_PROPERTIES = {
    MODEL_HUMIDIFIER_MJJSQ: MODEL_HUMIDIFIER_JSQ_COMMON,
    MODEL_HUMIDIFIER_JSQ: MODEL_HUMIDIFIER_JSQ_COMMON,
    MODEL_HUMIDIFIER_JSQ1: MODEL_HUMIDIFIER_JSQ_COMMON + ["wet_and_protect"],
}


class AirHumidifierException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Low = 1
    Medium = 2
    High = 3
    Humidity = 4
    WetAndProtect = 5


class AirHumidifierStatus(DeviceStatus):
    """Container for status reports from the air humidifier mjjsq."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Response of a Air Humidifier (deerma.humidifier.mjjsq):

        {'Humidifier_Gear': 4, 'Humidity_Value': 44, 'HumiSet_Value': 54,
         'Led_State': 1, 'OnOff_State': 0, 'TemperatureValue': 21,
         'TipSound_State': 1, 'waterstatus': 1, 'watertankstatus': 1}
        """

        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return "on" if self.data["OnOff_State"] == 1 else "off"

    @property
    def is_on(self) -> bool:
        """True if device is turned on."""
        return self.power == "on"

    @property
    def mode(self) -> OperationMode:
        """Operation mode.

        Can be either low, medium, high or humidity.
        """
        return OperationMode(self.data["Humidifier_Gear"])

    @property
    def temperature(self) -> int:
        """Current temperature in degree celsius."""
        return self.data["TemperatureValue"]

    @property
    def humidity(self) -> int:
        """Current humidity in percent."""
        return self.data["Humidity_Value"]

    @property
    def buzzer(self) -> bool:
        """True if buzzer is turned on."""
        return self.data["TipSound_State"] == 1

    @property
    def led(self) -> bool:
        """True if LED is turned on."""
        return self.data["Led_State"] == 1

    @property
    def target_humidity(self) -> int:
        """Target humiditiy in percent."""
        return self.data["HumiSet_Value"]

    @property
    def no_water(self) -> bool:
        """True if the water tank is empty."""
        return self.data["waterstatus"] == 0

    @property
    def water_tank_detached(self) -> bool:
        """True if the water tank is detached."""
        return self.data["watertankstatus"] == 0

    @property
    def wet_protection(self) -> Optional[bool]:
        """True if wet protection is enabled."""
        if self.data["wet_and_protect"] is not None:
            return self.data["wet_and_protect"] == 1

        return None


class AirHumidifierMjjsq(Device):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_HUMIDIFIER_MJJSQ,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_HUMIDIFIER_MJJSQ

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "Temperature: {result.temperature} °C\n"
            "Humidity: {result.humidity} %\n"
            "LED: {result.led}\n"
            "Buzzer: {result.buzzer}\n"
            "Target humidity: {result.target_humidity} %\n"
            "No water: {result.no_water}\n"
            "Water tank detached: {result.water_tank_detached}\n"
            "Wet protection: {result.wet_protection}\n",
        )
    )
    def status(self) -> AirHumidifierStatus:
        """Retrieve properties."""

        properties = AVAILABLE_PROPERTIES[self.model]
        values = self.get_properties(properties, max_properties=1)

        return AirHumidifierStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("Set_OnOff", [1])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("Set_OnOff", [0])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        return self.send("Set_HumidifierGears", [mode.value])

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        return self.send("SetLedState", [int(led)])

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        return self.send("SetTipSound_Status", [int(buzzer)])

    @command(
        click.argument("humidity", type=int),
        default_output=format_output("Setting target humidity to {humidity}"),
    )
    def set_target_humidity(self, humidity: int):
        """Set the target humidity in percent."""
        if humidity < 0 or humidity > 99:
            raise AirHumidifierException("Invalid target humidity: %s" % humidity)

        return self.send("Set_HumiValue", [humidity])

    @command(
        click.argument("protection", type=bool),
        default_output=format_output(
            lambda protection: "Turning on wet protection"
            if protection
            else "Turning off wet protection"
        ),
    )
    def set_wet_protection(self, protection: bool):
        """Turn wet protection on/off."""
        return self.send("Set_wet_and_protect", [int(protection)])



from unittest import TestCase

import pytest

from miio import Ceil
from miio.ceil import CeilException, CeilStatus

from .dummies import DummyDevice


class DummyCeil(DummyDevice, Ceil):
    def __init__(self, *args, **kwargs):
        self.state = {
            "power": "on",
            "bright": 50,
            "snm": 4,
            "dv": 0,
            "cctsw": [[0, 3], [0, 2], [0, 1]],
            "bl": 1,
            "mb": 1,
            "ac": 1,
            "mssw": 1,
            "cct": 99,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_bright": lambda x: self._set_state("bright", x),
            "apply_fixed_scene": lambda x: self._set_state("snm", x),
            "delay_off": lambda x: self._set_state("dv", x),
            "enable_bl": lambda x: self._set_state("bl", x),
            "enable_ac": lambda x: self._set_state("ac", x),
            "set_cct": lambda x: self._set_state("cct", x),
            "set_bricct": lambda x: (
                self._set_state("bright", [x[0]]),
                self._set_state("cct", [x[1]]),
            ),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def ceil(request):
    request.cls.device = DummyCeil()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("ceil")
class TestCeil(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(CeilStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().brightness == self.device.start_state["bright"]
        assert self.state().color_temperature == self.device.start_state["cct"]
        assert self.state().scene == self.device.start_state["snm"]
        assert self.state().delay_off_countdown == self.device.start_state["dv"]
        assert self.state().smart_night_light is (self.device.start_state["bl"] == 1)
        assert self.state().automatic_color_temperature is (
            self.device.start_state["ac"] == 1
        )

    def test_set_brightness(self):
        def brightness():
            return self.device.status().brightness

        self.device.set_brightness(10)
        assert brightness() == 10
        self.device.set_brightness(20)
        assert brightness() == 20

        with pytest.raises(CeilException):
            self.device.set_brightness(-1)

        with pytest.raises(CeilException):
            self.device.set_brightness(101)

    def test_set_color_temperature(self):
        def color_temperature():
            return self.device.status().color_temperature

        self.device.set_color_temperature(30)
        assert color_temperature() == 30
        self.device.set_color_temperature(20)
        assert color_temperature() == 20

        with pytest.raises(CeilException):
            self.device.set_color_temperature(-1)

        with pytest.raises(CeilException):
            self.device.set_color_temperature(101)

    def test_set_brightness_and_color_temperature(self):
        def color_temperature():
            return self.device.status().color_temperature

        def brightness():
            return self.device.status().brightness

        self.device.set_brightness_and_color_temperature(20, 21)
        assert brightness() == 20
        assert color_temperature() == 21
        self.device.set_brightness_and_color_temperature(31, 30)
        assert brightness() == 31
        assert color_temperature() == 30
        self.device.set_brightness_and_color_temperature(10, 11)
        assert brightness() == 10
        assert color_temperature() == 11

        with pytest.raises(CeilException):
            self.device.set_brightness_and_color_temperature(-1, 10)

        with pytest.raises(CeilException):
            self.device.set_brightness_and_color_temperature(10, -1)

        with pytest.raises(CeilException):
            self.device.set_brightness_and_color_temperature(0, 10)

        with pytest.raises(CeilException):
            self.device.set_brightness_and_color_temperature(10, 0)

        with pytest.raises(CeilException):
            self.device.set_brightness_and_color_temperature(101, 10)

        with pytest.raises(CeilException):
            self.device.set_brightness_and_color_temperature(10, 101)

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200

        with pytest.raises(CeilException):
            self.device.delay_off(0)

        with pytest.raises(CeilException):
            self.device.delay_off(-1)

    def test_set_scene(self):
        def scene():
            return self.device.status().scene

        self.device.set_scene(1)
        assert scene() == 1
        self.device.set_scene(4)
        assert scene() == 4

        with pytest.raises(CeilException):
            self.device.set_scene(0)

        with pytest.raises(CeilException):
            self.device.set_scene(5)

    def test_smart_night_light_on(self):
        def smart_night_light():
            return self.device.status().smart_night_light

        self.device.smart_night_light_off()
        assert smart_night_light() is False
        self.device.smart_night_light_on()
        assert smart_night_light() is True

    def test_automatic_color_temperature_on(self):
        def automatic_color_temperature():
            return self.device.status().automatic_color_temperature

        self.device.automatic_color_temperature_on()
        assert automatic_color_temperature() is True
        self.device.automatic_color_temperature_off()
        assert automatic_color_temperature() is False



from typing import Dict, Optional


class DeviceInfo:
    """Container of miIO device information.

    Hardware properties such as device model, MAC address, memory information, and
    hardware and software information is contained here.
    """

    def __init__(self, data):
        """Response of a Xiaomi Smart WiFi Plug.

        {'ap': {'bssid': 'FF:FF:FF:FF:FF:FF', 'rssi': -68, 'ssid': 'network'},
         'cfg_time': 0,
         'fw_ver': '1.2.4_16',
         'hw_ver': 'MW300',
         'life': 24,
         'mac': '28:FF:FF:FF:FF:FF',
         'mmfree': 30312,
         'model': 'chuangmi.plug.m1',
         'netif': {'gw': '192.168.xxx.x',
                   'localIp': '192.168.xxx.x',
                   'mask': '255.255.255.0'},
         'ot': 'otu',
         'ott_stat': [0, 0, 0, 0],
         'otu_stat': [320, 267, 3, 0, 3, 742],
         'token': '2b00042f7481c7b056c4b410d28f33cf',
         'wifi_fw_ver': 'SD878x-14.76.36.p84-702.1.0-WM'}
        """
        self.data = data

    def __repr__(self):
        return "%s v%s (%s) @ %s - token: %s" % (
            self.model,
            self.firmware_version,
            self.mac_address,
            self.ip_address,
            self.token,
        )

    @property
    def network_interface(self) -> Dict:
        """Information about network configuration.

        If unavailable, returns an empty dictionary.
        """
        return self.data.get("netif", {})

    @property
    def accesspoint(self):
        """Information about connected wlan accesspoint.

        If unavailable, returns an empty dictionary.
        """
        return self.data.get("ap", {})

    @property
    def model(self) -> Optional[str]:
        """Model string if available."""
        return self.data.get("model")

    @property
    def firmware_version(self) -> Optional[str]:
        """Firmware version if available."""
        return self.data.get("fw_ver")

    @property
    def hardware_version(self) -> Optional[str]:
        """Hardware version if available."""
        return self.data.get("hw_ver")

    @property
    def mac_address(self) -> Optional[str]:
        """MAC address, if available."""
        return self.data.get("mac")

    @property
    def ip_address(self) -> Optional[str]:
        """IP address, if available."""
        return self.network_interface.get("localIp")

    @property
    def token(self) -> Optional[str]:
        """Return the current device token."""
        return self.data.get("token")

    @property
    def raw(self):
        """Raw data as returned by the device."""
        return self.data



import enum
import logging

import click

from .click_common import command, format_output
from .exceptions import DeviceException
from .miot_device import DeviceStatus, MiotDevice

_LOGGER = logging.getLogger(__name__)

MODEL_AIRQUALITYMONITOR_CGDN1 = "cgllc.airm.cgdn1"

_MAPPING_CGDN1 = {
    # Source https://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:air-monitor:0000A008:cgllc-cgdn1:1
    # Environment
    "humidity": {"siid": 3, "piid": 1},  # [0, 100] step 1
    "pm25": {"siid": 3, "piid": 4},  # [0, 1000] step 1
    "pm10": {"siid": 3, "piid": 5},  # [0, 1000] step 1
    "temperature": {"siid": 3, "piid": 7},  # [-30, 100] step 0.00001
    "co2": {"siid": 3, "piid": 8},  # [0, 9999] step 1
    # Battery
    "battery": {"siid": 4, "piid": 1},  # [0, 100] step 1
    "charging_state": {
        "siid": 4,
        "piid": 2,
    },  # 1 - Charging, 2 - Not charging, 3 - Not chargeable
    "voltage": {"siid": 4, "piid": 3},  # [0, 65535] step 1
    # Settings
    "start_time": {"siid": 9, "piid": 2},  # [0, 2147483647] step 1
    "end_time": {"siid": 9, "piid": 3},  # [0, 2147483647] step 1
    "monitoring_frequency": {
        "siid": 9,
        "piid": 4,
    },  # 1, 60, 300, 600, 0; device accepts [0..600]
    "screen_off": {
        "siid": 9,
        "piid": 5,
    },  # 15, 30, 60, 300, 0; device accepts [0..300], 0 means never
    "device_off": {
        "siid": 9,
        "piid": 6,
    },  # 15, 30, 60, 0; device accepts [0..60], 0 means never
    "temperature_unit": {"siid": 9, "piid": 7},
}


class AirQualityMonitorMiotException(DeviceException):
    pass


class ChargingState(enum.Enum):
    Unplugged = 0  # Not mentioned in the spec
    Charging = 1
    NotCharging = 2
    NotChargable = 3


class MonitoringFrequencyCGDN1(enum.Enum):  # Official spec options
    Every1Second = 1
    Every1Minute = 60
    Every5Minutes = 300
    Every10Minutes = 600
    NotSet = 0


class ScreenOffCGDN1(enum.Enum):  # Official spec options
    After15Seconds = 15
    After30Seconds = 30
    After1Minute = 60
    After5Minutes = 300
    Never = 0


class DeviceOffCGDN1(enum.Enum):  # Official spec options
    After15Minutes = 15
    After30Minutes = 30
    After1Hour = 60
    Never = 0


class DisplayTemperatureUnitCGDN1(enum.Enum):
    Celcius = "c"
    Fahrenheit = "f"


class AirQualityMonitorCGDN1Status(DeviceStatus):
    """
    Container of air quality monitor CGDN1 status.

    {
      'humidity': 34,
      'pm25': 18,
      'pm10': 21,
      'temperature': 22.8,
      'co2': 468,
      'battery': 37,
      'charging_state': 0,
      'voltage': 3564,
      'start_time': 0,
      'end_time': 0,
      'monitoring_frequency': 1,
      'screen_off': 300,
      'device_off': 60,
      'temperature_unit': 'c'
    }

    """

    def __init__(self, data):
        self.data = data

    @property
    def humidity(self) -> int:
        """Return humidity value (0...100%)."""
        return self.data["humidity"]

    @property
    def pm25(self) -> int:
        """Return PM 2.5 value (0...1000ppm)."""
        return self.data["pm25"]

    @property
    def pm10(self) -> int:
        """Return PM 10 value (0...1000ppm)."""
        return self.data["pm10"]

    @property
    def temperature(self) -> float:
        """Return temperature value (-30...100°C)."""
        return self.data["temperature"]

    @property
    def co2(self) -> int:
        """Return co2 value (0...9999ppm)."""
        return self.data["co2"]

    @property
    def battery(self) -> int:
        """Return battery level (0...100%)."""
        return self.data["battery"]

    @property
    def charging_state(self) -> ChargingState:
        """Return charging state."""
        return ChargingState(self.data["charging_state"])

    @property
    def monitoring_frequency(self) -> int:
        """Return monitoring frequency time (0..600 s)."""
        return self.data["monitoring_frequency"]

    @property
    def screen_off(self) -> int:
        """Return screen off time (0..300 s)."""
        return self.data["screen_off"]

    @property
    def device_off(self) -> int:
        """Return device off time (0..60 min)."""
        return self.data["device_off"]

    @property
    def display_temperature_unit(self):
        """Return display temperature unit."""
        return DisplayTemperatureUnitCGDN1(self.data["temperature_unit"])


class AirQualityMonitorCGDN1(MiotDevice):
    """Qingping Air Monitor Lite."""

    mapping = _MAPPING_CGDN1

    @command(
        default_output=format_output(
            "",
            "Humidity: {result.humidity} %\n"
            "PM 2.5: {result.pm25} μg/m³\n"
            "PM 10: {result.pm10} μg/m³\n"
            "Temperature: {result.temperature} °C\n"
            "CO₂: {result.co2} μg/m³\n"
            "Battery: {result.battery} %\n"
            "Charging state: {result.charging_state.name}\n"
            "Monitoring frequency: {result.monitoring_frequency} s\n"
            "Screen off: {result.screen_off} s\n"
            "Device off: {result.device_off} min\n"
            "Display temperature unit: {result.display_temperature_unit.name}\n",
        )
    )
    def status(self) -> AirQualityMonitorCGDN1Status:
        """Retrieve properties."""

        return AirQualityMonitorCGDN1Status(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping()
            }
        )

    @command(
        click.argument("duration", type=int),
        default_output=format_output("Setting monitoring frequency to {duration} s"),
    )
    def set_monitoring_frequency_duration(self, duration):
        """Set monitoring frequency."""
        if duration < 0 or duration > 600:
            raise AirQualityMonitorMiotException(
                "Invalid duration: %s. Must be between 0 and 600" % duration
            )
        return self.set_property("monitoring_frequency", duration)

    @command(
        click.argument("duration", type=int),
        default_output=format_output("Setting device off duration to {duration} min"),
    )
    def set_device_off_duration(self, duration):
        """Set device off duration."""
        if duration < 0 or duration > 60:
            raise AirQualityMonitorMiotException(
                "Invalid duration: %s. Must be between 0 and 60" % duration
            )
        return self.set_property("device_off", duration)

    @command(
        click.argument("duration", type=int),
        default_output=format_output("Setting screen off duration to {duration} s"),
    )
    def set_screen_off_duration(self, duration):
        """Set screen off duration."""
        if duration < 0 or duration > 300:
            raise AirQualityMonitorMiotException(
                "Invalid duration: %s. Must be between 0 and 300" % duration
            )
        return self.set_property("screen_off", duration)

    @command(
        click.argument(
            "unit",
            type=click.Choice(DisplayTemperatureUnitCGDN1.__members__),
            callback=lambda c, p, v: getattr(DisplayTemperatureUnitCGDN1, v),
        ),
        default_output=format_output("Setting display temperature unit to {unit.name}"),
    )
    def set_display_temperature_unit(self, unit: DisplayTemperatureUnitCGDN1):
        """Set display temperature unit."""
        return self.set_property("temperature_unit", unit.value)



from unittest import TestCase

import pytest

from miio import AirFresh
from miio.airfresh import (
    MODEL_AIRFRESH_VA2,
    MODEL_AIRFRESH_VA4,
    AirFreshException,
    AirFreshStatus,
    LedBrightness,
    OperationMode,
)

from .dummies import DummyDevice


class DummyAirFresh(DummyDevice, AirFresh):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_AIRFRESH_VA2
        self.state = {
            "power": "on",
            "ptc_state": None,
            "temp_dec": 186,
            "aqi": 10,
            "average_aqi": 8,
            "humidity": 62,
            "co2": 350,
            "buzzer": "off",
            "child_lock": "off",
            "led_level": 2,
            "mode": "auto",
            "motor1_speed": 354,
            "use_time": 2457000,
            "ntcT": None,
            "app_extra": 1,
            "f1_hour_used": 682,
            "filter_life": 80,
            "f_hour": 3500,
            "favorite_level": None,
            "led": "on",
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_led": lambda x: self._set_state("led", x),
            "set_led_level": lambda x: self._set_state("led_level", x),
            "reset_filter1": lambda x: (
                self._set_state("f1_hour_used", [0]),
                self._set_state("filter_life", [100]),
            ),
            "set_app_extra": lambda x: self._set_state("app_extra", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def airfresh(request):
    request.cls.device = DummyAirFresh()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airfresh")
class TestAirFresh(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            AirFreshStatus(self.device.start_state, MODEL_AIRFRESH_VA2)
        )

        assert self.is_on() is True
        assert self.state().ptc is None
        assert self.state().aqi == self.device.start_state["aqi"]
        assert self.state().average_aqi == self.device.start_state["average_aqi"]
        assert self.state().temperature == self.device.start_state["temp_dec"] / 10.0
        assert self.state().ntc_temperature is None
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().co2 == self.device.start_state["co2"]
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert (
            self.state().filter_life_remaining == self.device.start_state["filter_life"]
        )
        assert self.state().filter_hours_used == self.device.start_state["f1_hour_used"]
        assert self.state().use_time == self.device.start_state["use_time"]
        assert self.state().motor_speed == self.device.start_state["motor1_speed"]
        assert self.state().led == (self.device.start_state["led"] == "on")
        assert self.state().led_brightness == LedBrightness(
            self.device.start_state["led_level"]
        )
        assert self.state().buzzer == (self.device.start_state["buzzer"] == "on")
        assert self.state().child_lock == (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().extra_features == self.device.start_state["app_extra"]

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Auto)
        assert mode() == OperationMode.Auto

        self.device.set_mode(OperationMode.Silent)
        assert mode() == OperationMode.Silent

        self.device.set_mode(OperationMode.Interval)
        assert mode() == OperationMode.Interval

        self.device.set_mode(OperationMode.Low)
        assert mode() == OperationMode.Low

        self.device.set_mode(OperationMode.Middle)
        assert mode() == OperationMode.Middle

        self.device.set_mode(OperationMode.Strong)
        assert mode() == OperationMode.Strong

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_set_extra_features(self):
        def extra_features():
            return self.device.status().extra_features

        self.device.set_extra_features(0)
        assert extra_features() == 0
        self.device.set_extra_features(1)
        assert extra_features() == 1
        self.device.set_extra_features(2)
        assert extra_features() == 2

        with pytest.raises(AirFreshException):
            self.device.set_extra_features(-1)

    def test_reset_filter(self):
        def filter_hours_used():
            return self.device.status().filter_hours_used

        def filter_life_remaining():
            return self.device.status().filter_life_remaining

        self.device._reset_state()
        assert filter_hours_used() != 0
        assert filter_life_remaining() != 100
        self.device.reset_filter()
        assert filter_hours_used() == 0
        assert filter_life_remaining() == 100


class DummyAirFreshVA4(DummyDevice, AirFresh):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_AIRFRESH_VA4
        self.state = {
            "power": "on",
            "ptc_state": "off",
            "temp_dec": 18.6,
            "aqi": 10,
            "average_aqi": 8,
            "humidity": 62,
            "co2": 350,
            "buzzer": "off",
            "child_lock": "off",
            "led_level": 2,
            "mode": "auto",
            "motor1_speed": 354,
            "use_time": 2457000,
            "ntcT": 33.53,
            "app_extra": 1,
            "f1_hour_used": 682,
            "filter_life": 80,
            "f_hour": 3500,
            "favorite_level": None,
            "led": "on",
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_ptc_state": lambda x: self._set_state("ptc_state", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_led": lambda x: self._set_state("led", x),
            "set_led_level": lambda x: self._set_state("led_level", x),
            "reset_filter1": lambda x: (
                self._set_state("f1_hour_used", [0]),
                self._set_state("filter_life", [100]),
            ),
            "set_app_extra": lambda x: self._set_state("app_extra", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def airfreshva4(request):
    request.cls.device = DummyAirFreshVA4()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airfreshva4")
class TestAirFreshVA4(TestAirFresh):
    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(
            AirFreshStatus(self.device.start_state, MODEL_AIRFRESH_VA4)
        )

        assert self.is_on() is True
        assert self.state().ptc == (self.device.start_state["ptc_state"] == "on")
        assert self.state().aqi == self.device.start_state["aqi"]
        assert self.state().average_aqi == self.device.start_state["average_aqi"]
        assert self.state().temperature == self.device.start_state["temp_dec"]
        assert self.state().ntc_temperature == self.device.start_state["ntcT"]
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().co2 == self.device.start_state["co2"]
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert (
            self.state().filter_life_remaining == self.device.start_state["filter_life"]
        )
        assert self.state().filter_hours_used == self.device.start_state["f1_hour_used"]
        assert self.state().use_time == self.device.start_state["use_time"]
        assert self.state().motor_speed == self.device.start_state["motor1_speed"]
        assert self.state().led == (self.device.start_state["led"] == "on")
        assert self.state().led_brightness == LedBrightness(
            self.device.start_state["led_level"]
        )
        assert self.state().buzzer == (self.device.start_state["buzzer"] == "on")
        assert self.state().child_lock == (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().extra_features == self.device.start_state["app_extra"]

    def test_set_ptc(self):
        def ptc():
            return self.device.status().ptc

        self.device.set_ptc(True)
        assert ptc() is True

        self.device.set_ptc(False)
        assert ptc() is False



import enum
import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceInfo, DeviceStatus
from .exceptions import DeviceError, DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_DEHUMIDIFIER_V1 = "nwt.derh.wdh318efw1"

AVAILABLE_PROPERTIES = {
    MODEL_DEHUMIDIFIER_V1: [
        "on_off",
        "mode",
        "fan_st",
        "buzzer",
        "led",
        "child_lock",
        "humidity",
        "temp",
        "compressor_status",
        "fan_speed",
        "tank_full",
        "defrost_status",
        "alarm",
        "auto",
    ]
}


class AirDehumidifierException(DeviceException):
    pass


class OperationMode(enum.Enum):
    On = "on"
    Auto = "auto"
    DryCloth = "dry_cloth"


class FanSpeed(enum.Enum):
    Sleep = 0
    Low = 1
    Medium = 2
    High = 3
    Strong = 4


class AirDehumidifierStatus(DeviceStatus):
    """Container for status reports from the air dehumidifier."""

    def __init__(self, data: Dict[str, Any], device_info: DeviceInfo) -> None:
        """Response of a Air Dehumidifier (nwt.derh.wdh318efw1):

        {'on_off': 'on', 'mode': 'auto', 'fan_st': 2,
         'buzzer': 'off', 'led': 'on', 'child_lock': 'off',
         'humidity': 47, 'temp': 34, 'compressor_status': 'off',
         'fan_speed': 0, 'tank_full': 'off', 'defrost_status': 'off,
         'alarm': 'ok','auto': 50}
        """

        self.data = data
        self.device_info = device_info

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["on_off"]

    @property
    def is_on(self) -> bool:
        """True if device is turned on."""
        return self.power == "on"

    @property
    def mode(self) -> OperationMode:
        """Operation mode.

        Can be either on, auth or dry_cloth.
        """
        return OperationMode(self.data["mode"])

    @property
    def temperature(self) -> Optional[float]:
        """Current temperature, if available."""
        if "temp" in self.data and self.data["temp"] is not None:
            return self.data["temp"]
        return None

    @property
    def humidity(self) -> int:
        """Current humidity."""
        return self.data["humidity"]

    @property
    def buzzer(self) -> bool:
        """True if buzzer is turned on."""
        return self.data["buzzer"] == "on"

    @property
    def led(self) -> bool:
        """LED brightness if available."""
        return self.data["led"] == "on"

    @property
    def child_lock(self) -> bool:
        """Return True if child lock is on."""
        return self.data["child_lock"] == "on"

    @property
    def target_humidity(self) -> Optional[int]:
        """Target humiditiy.

        Can be either 40, 50, 60 percent.
        """
        if "auto" in self.data and self.data["auto"] is not None:
            return self.data["auto"]
        return None

    @property
    def fan_speed(self) -> Optional[FanSpeed]:
        """Current fan speed."""
        if "fan_speed" in self.data and self.data["fan_speed"] is not None:
            return FanSpeed(self.data["fan_speed"])
        return None

    @property
    def tank_full(self) -> bool:
        """The remaining amount of water in percent."""
        return self.data["tank_full"] == "on"

    @property
    def compressor_status(self) -> bool:
        """Compressor status."""
        return self.data["compressor_status"] == "on"

    @property
    def defrost_status(self) -> bool:
        """Defrost status."""
        return self.data["defrost_status"] == "on"

    @property
    def fan_st(self) -> int:
        """Fan st."""
        return self.data["fan_st"]

    @property
    def alarm(self) -> str:
        """Alarm."""
        return self.data["alarm"]


class AirDehumidifier(Device):
    """Implementation of Xiaomi Mi Air Dehumidifier."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_DEHUMIDIFIER_V1,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_DEHUMIDIFIER_V1

        self.device_info: DeviceInfo

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "Temperature: {result.temperature} °C\n"
            "Humidity: {result.humidity} %\n"
            "Buzzer: {result.buzzer}\n"
            "LED : {result.led}\n"
            "Child lock: {result.child_lock}\n"
            "Target humidity: {result.target_humidity} %\n"
            "Fan speed: {result.fan_speed}\n"
            "Tank Full: {result.tank_full}\n"
            "Compressor Status: {result.compressor_status}\n"
            "Defrost Status: {result.defrost_status}\n"
            "Fan st: {result.fan_st}\n"
            "Alarm: {result.alarm}\n",
        )
    )
    def status(self) -> AirDehumidifierStatus:
        """Retrieve properties."""

        if self.device_info is None:
            self.device_info = self.info()

        properties = AVAILABLE_PROPERTIES[self.model]

        values = self.get_properties(properties, max_properties=1)

        return AirDehumidifierStatus(
            defaultdict(lambda: None, zip(properties, values)), self.device_info
        )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        try:
            return self.send("set_mode", [mode.value])
        except DeviceError as error:
            # {'code': -6011, 'message': 'device_poweroff'}
            if error.code == -6011:
                self.on()
                return self.send("set_mode", [mode.value])
            raise

    @command(
        click.argument("fan_speed", type=EnumType(FanSpeed)),
        default_output=format_output("Setting fan level to {fan_speed}"),
    )
    def set_fan_speed(self, fan_speed: FanSpeed):
        """Set the fan speed."""
        try:
            return self.send("set_fan_level", [fan_speed.value])
        except DeviceError as ex:
            if ex.code == -10000:
                raise AirDehumidifierException(
                    "Unable to set fan speed, this can happen if device is turned off."
                ) from ex

            raise

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        if led:
            return self.send("set_led", ["on"])
        else:
            return self.send("set_led", ["off"])

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        if buzzer:
            return self.send("set_buzzer", ["on"])
        else:
            return self.send("set_buzzer", ["off"])

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        if lock:
            return self.send("set_child_lock", ["on"])
        else:
            return self.send("set_child_lock", ["off"])

    @command(
        click.argument("humidity", type=int),
        default_output=format_output("Setting target humidity to {humidity}"),
    )
    def set_target_humidity(self, humidity: int):
        """Set the auto target humidity."""
        if humidity not in [40, 50, 60]:
            raise AirDehumidifierException(
                "Invalid auto target humidity: %s" % humidity
            )

        return self.send("set_auto", [humidity])



from datetime import timedelta
from unittest import TestCase

import pytest

from miio import RoidmiVacuumMiot
from miio.roidmivacuum_miot import (
    ChargingState,
    FanSpeed,
    PathMode,
    RoidmiState,
    SweepMode,
    SweepType,
    WaterLevel,
)
from miio.vacuumcontainers import DNDStatus

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "auto_boost": 1,
    "battery_level": 42,
    "main_brush_life_level": 85,
    "side_brushes_life_level": 57,
    "sensor_dirty_remaning_level": 60,
    "main_brush_left_minutes": 235,
    "side_brushes_left_minutes": 187,
    "sensor_dirty_time_left_minutes": 1096,
    "charging_state": ChargingState.Charging,
    "fanspeed_mode": FanSpeed.FullSpeed,
    "current_audio": "girl_en",
    "clean_area": 27,
    "error_code": 0,
    "state": RoidmiState.Paused.value,
    "double_clean": 0,
    "filter_left_minutes": 154,
    "filter_life_level": 66,
    "forbid_mode": '{"time":[75600,21600,1],"tz":2,"tzs":7200}',
    "led_switch": 0,
    "lidar_collision": 1,
    "mop_present": 1,
    "mute": 0,
    "station_key": 0,
    "station_led": 0,
    # "station_type": {"siid": 8, "piid": 29}, # uint32
    # "switch_status": {"siid": 2, "piid": 10},
    "sweep_mode": SweepMode.Smart,
    "sweep_type": SweepType.MopAndSweep,
    "timing": '{"time":[[32400,1,3,0,[1,2,3,4,5],0,[12,10],null],[57600,0,1,2,[1,2,3,4,5,6,0],2,[],null]],"tz":2,"tzs":7200}',
    "path_mode": PathMode.Normal,
    "work_station_freq": 1,
    # "uid": "12345678",
    "volume": 4,
    "water_level": WaterLevel.Mop,
    "total_clean_time_sec": 321456,
    "total_clean_areas": 345678,
    "clean_counts": 987,
    "clean_time_sec": 32,
}


class DummyRoidmiVacuumMiot(DummyMiotDevice, RoidmiVacuumMiot):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def dummyroidmivacuum(request):
    request.cls.device = DummyRoidmiVacuumMiot()


@pytest.mark.usefixtures("dummyroidmivacuum")
class TestRoidmiVacuum(TestCase):
    def test_vacuum_status(self):
        status = self.device.status()
        assert status.carpet_mode == _INITIAL_STATE["auto_boost"]
        assert status.battery == _INITIAL_STATE["battery_level"]
        assert status.charging_state == ChargingState(_INITIAL_STATE["charging_state"])
        assert status.fan_speed == FanSpeed(_INITIAL_STATE["fanspeed_mode"])
        assert status.current_audio == _INITIAL_STATE["current_audio"]
        assert status.clean_area == _INITIAL_STATE["clean_area"]
        assert status.clean_time.total_seconds() == _INITIAL_STATE["clean_time_sec"]
        assert status.error_code == _INITIAL_STATE["error_code"]
        assert status.error == "NoFaults"
        assert status.state == RoidmiState(_INITIAL_STATE["state"])
        assert status.double_clean == _INITIAL_STATE["double_clean"]
        assert str(status.dnd_status) == str(
            status._parse_forbid_mode(_INITIAL_STATE["forbid_mode"])
        )
        assert status.led == _INITIAL_STATE["led_switch"]
        assert status.is_lidar_collision_sensor == _INITIAL_STATE["lidar_collision"]
        assert status.is_mop_attached == _INITIAL_STATE["mop_present"]
        assert status.is_muted == _INITIAL_STATE["mute"]
        assert status.station_key == _INITIAL_STATE["station_key"]
        assert status.station_led == _INITIAL_STATE["station_led"]
        assert status.sweep_mode == SweepMode(_INITIAL_STATE["sweep_mode"])
        assert status.sweep_type == SweepType(_INITIAL_STATE["sweep_type"])
        assert status.timing == _INITIAL_STATE["timing"]
        assert status.path_mode == PathMode(_INITIAL_STATE["path_mode"])
        assert status.dust_collection_frequency == _INITIAL_STATE["work_station_freq"]
        assert status.volume == _INITIAL_STATE["volume"]
        assert status.water_level == WaterLevel(_INITIAL_STATE["water_level"])

        assert status.is_paused is True
        assert status.is_on is False
        assert status.got_error is False

    def test_cleaning_summary(self):
        status = self.device.cleaning_summary()
        assert (
            status.total_duration.total_seconds()
            == _INITIAL_STATE["total_clean_time_sec"]
        )
        assert status.total_area == _INITIAL_STATE["total_clean_areas"]
        assert status.count == _INITIAL_STATE["clean_counts"]

    def test_consumable_status(self):
        status = self.device.consumable_status()
        assert (
            status.main_brush_left.total_seconds() / 60
            == _INITIAL_STATE["main_brush_left_minutes"]
        )
        assert (
            status.side_brush_left.total_seconds() / 60
            == _INITIAL_STATE["side_brushes_left_minutes"]
        )
        assert (
            status.sensor_dirty_left.total_seconds() / 60
            == _INITIAL_STATE["sensor_dirty_time_left_minutes"]
        )
        assert status.main_brush == status._calcUsageTime(
            status.main_brush_left, _INITIAL_STATE["main_brush_life_level"]
        )
        assert status.side_brush == status._calcUsageTime(
            status.side_brush_left, _INITIAL_STATE["side_brushes_life_level"]
        )
        assert status.sensor_dirty == status._calcUsageTime(
            status.sensor_dirty_left, _INITIAL_STATE["sensor_dirty_remaning_level"]
        )
        assert (
            status.filter_left.total_seconds() / 60
            == _INITIAL_STATE["filter_left_minutes"]
        )
        assert status.filter == status._calcUsageTime(
            status.filter_left, _INITIAL_STATE["filter_life_level"]
        )

    def test__calcUsageTime(self):
        status = self.device.consumable_status()
        orig_time = timedelta(minutes=500)
        remaning_level = 30
        remaning_time = orig_time * 0.30
        used_time = orig_time - remaning_time
        assert used_time == status._calcUsageTime(remaning_time, remaning_level)

    def test_parse_forbid_mode(self):
        status = self.device.status()
        value = '{"time":[75600,21600,1],"tz":2,"tzs":7200}'
        expected_value = DNDStatus(
            dict(
                enabled=True,
                start_hour=21,
                start_minute=0,
                end_hour=6,
                end_minute=0,
            )
        )
        assert str(status._parse_forbid_mode(value)) == str(expected_value)

    def test_parse_forbid_mode2(self):
        status = self.device.status()
        value = '{"time":[82080,33300,0],"tz":3,"tzs":10800}'
        expected_value = DNDStatus(
            dict(
                enabled=False,
                start_hour=22,
                start_minute=48,
                end_hour=9,
                end_minute=15,
            )
        )
        assert str(status._parse_forbid_mode(value)) == str(expected_value)


class DummyRoidmiVacuumMiot2(DummyMiotDevice, RoidmiVacuumMiot):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        self.state["charging_state"] = -10
        self.state["fanspeed_mode"] = -11
        self.state["state"] = -12
        self.state["sweep_mode"] = -13
        self.state["sweep_type"] = -14
        self.state["path_mode"] = -15
        self.state["water_level"] = -16
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def dummyroidmivacuum2(request):
    request.cls.device = DummyRoidmiVacuumMiot2()


@pytest.mark.usefixtures("dummyroidmivacuum2")
class TestRoidmiVacuum2(TestCase):
    def test_vacuum_status_unexpected_values(self):
        status = self.device.status()
        assert status.charging_state == ChargingState.Unknown
        assert status.fan_speed == FanSpeed.Unknown
        assert status.state == RoidmiState.Unknown
        assert status.sweep_mode == SweepMode.Unknown
        assert status.sweep_type == SweepType.Unknown
        assert status.path_mode == PathMode.Unknown
        assert status.water_level == WaterLevel.Unknown
